# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from cassandra.query import BatchStatement
from cassandra.query import SimpleStatement
from oslotest import base

from monasca_persister.repositories.cassandra import metric_batch


class FakeException(Exception):
    pass


class TestMetricBatch(base.BaseTestCase):
    def setUp(self):
        super(TestMetricBatch, self).setUp()
        self.bound_stmt = SimpleStatement("whatever")
        self.metric_batch = metric_batch.MetricBatch(mock.Mock(),
                                                     mock.Mock(), 1)
        self._set_patchers()
        self._set_mocks()

    def _set_patchers(self):
        self.patch_from_key = mock.patch.object(
            self.metric_batch.metadata.token_map.token_class, 'from_key')
        self.patch_lb_make_query_plan = mock.patch.object(
            self.metric_batch.lb_policy, 'make_query_plan')

    def _set_mocks(self):
        self.mock_key = self.patch_from_key.start()
        self.mock_lb_make_query = self.patch_lb_make_query_plan.start()
        self.mock_key.return_value = 'token'
        self.mock_lb_make_query.return_value = 'token'

    def tearDown(self):
        super(TestMetricBatch, self).tearDown()
        self.mock_key.reset_mock()
        self.patch_from_key.stop()
        self.mock_lb_make_query.reset_mock()
        self.patch_lb_make_query_plan.stop()

    def test_batch_query_by_token_creates_a_query_map_tuple(self):
        query_map = {}

        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        batch_statement, counter = query_map.get('token')[0]

        self.assertIsInstance(batch_statement, BatchStatement)
        self.assertIsInstance(counter, metric_batch.Counter)

    def test_batch_query_increments_counter_if_counter_value_lt_batch_limit(self):
        self.metric_batch.batch_limit = 2
        query_map = {}

        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        prev_counter = query_map.get('token')[0][1].value()

        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        current_counter = query_map.get('token')[0][1].value()

        self.assertEqual(prev_counter + 1, current_counter)

    def test_new_batch_added_if_counter_value_gt_or_eq_batch_limit(self):
        self.metric_batch.batch_limit = 1
        query_map = {}

        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        prev_len = len(query_map.get('token'))

        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        curr_len = len(query_map.get('token'))

        self.assertEqual(prev_len + 1, curr_len)

    def test_batch_query_by_replicas_creates_a_query_map_tuple(self):
        query_map = {}

        self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
        batch_statement, counter = query_map.get(tuple('token'))[0]

        self.assertIsInstance(batch_statement, BatchStatement)
        self.assertIsInstance(counter, metric_batch.Counter)

    def test_batch_query_by_replicas_increments_counter_if_counter_value_lt_30(self):
        query_map = {}
        self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
        prev_counter = query_map.get(tuple('token'))[0][1]._count = 29

        self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
        current_counter = query_map.get(tuple('token'))[0][1].value()

        self.assertEqual(prev_counter + 1, current_counter)

    def test_batch_query_by_replicas_adds_new_batch_if_counter_value_gt_30(self):
        query_map = {}
        self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
        prev_len = len(query_map.get(tuple('token')))

        query_map.get(tuple('token'))[0][1]._count = 30
        self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
        curr_len = len(query_map.get(tuple('token')))

        self.assertEqual(prev_len + 1, curr_len)

    def test_add_metric_query(self):
        self.metric_batch.add_metric_query(self.bound_stmt)

    def test_add_dimension_query(self):
        self.metric_batch.add_dimension_query(self.bound_stmt)

    def test_add_dimension_metric_query(self):
        self.metric_batch.add_dimension_metric_query(self.bound_stmt)

    def test_add_metric_dimension_query(self):
        self.metric_batch.add_metric_dimension_query(self.bound_stmt)

    def test_add_measurement_query(self):
        self.metric_batch.add_measurement_query(self.bound_stmt)

    def test_clear(self):
        self.metric_batch.clear()

        self.assertEqual(self.metric_batch.metric_queries, {})
        self.assertEqual(self.metric_batch.dimension_queries, {})
        self.assertEqual(self.metric_batch.dimension_metric_queries, {})
        self.assertEqual(self.metric_batch.metric_dimension_queries, {})
        self.assertEqual(self.metric_batch.measurement_queries, {})

    def test_log_token_batch_map(self):
        query_map = {}
        self.metric_batch.batch_query_by_token(self.bound_stmt, query_map)
        with mock.patch.object(metric_batch.LOG, 'info') as mock_log_info:
            self.metric_batch.log_token_batch_map('name', query_map)
            mock_log_info.assert_called_with('name : Size: 1;  Tokens: |token: 1|')

    def test_log_replica_batch_map(self):
        query_map = {}
        with mock.patch.object(self.metric_batch.lb_policy, 'make_query_plan'):
            with mock.patch.object(metric_batch.LOG, 'info') as mock_log_info:
                self.metric_batch.batch_query_by_replicas(self.bound_stmt, query_map)
                self.metric_batch.log_replica_batch_map('name', query_map)
                mock_log_info.assert_called_with('name : Size: 1;  Replicas: |: 1|')

    def test_get_all_batches(self):
        with mock.patch.object(self.metric_batch, 'log_token_batch_map'):
            with(mock.patch.object(self.metric_batch, 'log_replica_batch_map')):
                sample_elements = ['measurement_query', 'metric_query', 'dimension_query',
                                   'dimension_metric_query', 'metric_dimension_query']
                self.metric_batch.measurement_queries.update({'some': sample_elements[0]})
                self.metric_batch.metric_queries.update({'some': sample_elements[1]})
                self.metric_batch.dimension_queries.update({'some': sample_elements[2]})
                self.metric_batch.dimension_metric_queries.update({'some': sample_elements[3]})
                self.metric_batch.metric_dimension_queries.update({'some': sample_elements[4]})

                result_list = self.metric_batch.get_all_batches()

                for elem in sample_elements:
                    self.assertIn(elem[0], result_list)

    def test_counter_increment_by(self):
        counter = metric_batch.Counter()
        temp_counter_value = counter.value()
        counter.increment_by(5)

        self.assertEqual(temp_counter_value + 5, counter.value())



# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cassandra.query import BatchStatement
from cassandra.query import BatchType
from oslo_log import log

LOG = log.getLogger(__name__)


class MetricBatch(object):
    def __init__(self, metadata, load_balance_policy, batch_limit):

        self.metadata = metadata
        self.batch_limit = batch_limit
        self.lb_policy = load_balance_policy
        self.metric_queries = dict()
        self.dimension_queries = dict()
        self.dimension_metric_queries = dict()
        self.metric_dimension_queries = dict()
        self.measurement_queries = dict()

    def batch_query_by_token(self, bound_stmt, query_map):
        token = self.metadata.token_map.token_class.from_key(bound_stmt.routing_key)

        queue = query_map.get(token, None)
        if not queue:
            queue = []
            batch = BatchStatement(BatchType.UNLOGGED)
            batch.add(bound_stmt)
            queue.append((batch, Counter(1)))
            query_map[token] = queue
        else:
            (batch, counter) = queue[-1]
            if counter.value() < self.batch_limit:
                batch.add(bound_stmt)
                counter.increment()
            else:
                batch = BatchStatement(BatchType.UNLOGGED)
                batch.add(bound_stmt)
                queue.append((batch, Counter(1)))

    def add_metric_query(self, bound_stmt):
        self.batch_query_by_token(bound_stmt, self.metric_queries)

    def add_dimension_query(self, bound_stmt):
        self.batch_query_by_token(bound_stmt, self.dimension_queries)

    def add_dimension_metric_query(self, bound_stmt):
        self.batch_query_by_token(bound_stmt, self.dimension_metric_queries)

    def add_metric_dimension_query(self, bound_stmt):
        self.batch_query_by_token(bound_stmt, self.metric_dimension_queries)

    def add_measurement_query(self, bound_stmt):
        self.batch_query_by_replicas(bound_stmt, self.measurement_queries)

    def batch_query_by_replicas(self, bound_stmt, query_map):
        hosts = tuple(
            self.lb_policy.make_query_plan(
                working_keyspace=bound_stmt.keyspace,
                query=bound_stmt))

        queue = query_map.get(hosts, None)
        if not queue:
            queue = []
            batch = BatchStatement(BatchType.UNLOGGED)
            batch.add(bound_stmt)
            queue.append((batch, Counter(1)))
            query_map[hosts] = queue
        else:
            (batch, counter) = queue[-1]
            if counter.value() < 30:
                batch.add(bound_stmt)
                counter.increment()
            else:
                batch = BatchStatement(BatchType.UNLOGGED)
                batch.add(bound_stmt)
                queue.append((batch, Counter(1)))

    def clear(self):
        self.metric_queries.clear()
        self.dimension_queries.clear()
        self.dimension_metric_queries.clear()
        self.metric_dimension_queries.clear()
        self.measurement_queries.clear()

    @staticmethod
    def log_token_batch_map(name, query_map):
        LOG.info('%s : Size: %s;  Tokens: |%s|' %
                 (name, len(query_map),
                  '|'.join(['%s: %s' % (
                      token,
                      ','.join([str(counter.value()) for (batch, counter) in queue]))
                      for token, queue in query_map.items()])))

    @staticmethod
    def log_replica_batch_map(name, query_map):
        LOG.info('%s : Size: %s;  Replicas: |%s|' %
                 (name, len(query_map), '|'.join([
                     '%s: %s' % (
                         ','.join([h.address for h in hosts]),
                         ','.join([str(counter.value()) for (batch, counter) in queue]))
                     for hosts, queue in query_map.items()])))

    def get_all_batches(self):
        self.log_token_batch_map("metric batches", self.metric_queries)
        self.log_token_batch_map("dimension batches", self.dimension_queries)
        self.log_token_batch_map("dimension metric batches", self.dimension_metric_queries)
        self.log_token_batch_map("metric dimension batches", self.metric_dimension_queries)
        self.log_replica_batch_map("measurement batches", self.measurement_queries)

        result_list = []

        for q in self.measurement_queries.values():
            result_list.extend(q)

        for q in self.metric_queries.values():
            result_list.extend(q)

        for q in self.dimension_queries.values():
            result_list.extend(q)

        for q in self.dimension_metric_queries.values():
            result_list.extend(q)

        for q in self.metric_dimension_queries.values():
            result_list.extend(q)

        return result_list


class Counter(object):
    def __init__(self, init_value=0):
        self._count = init_value

    def increment(self):
        self._count += 1

    def increment_by(self, increment):
        self._count += increment

    def value(self):
        return self._count



# (C) Copyright 2014-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Persister Module

   The Persister reads metrics and alarms from Kafka and then stores them
   in into either Influxdb or Cassandra

   Start the perister as stand-alone process by running 'persister.py
   --config-file <config file>'
"""
import multiprocessing
import os
import signal
import sys
import time

from monasca_common.simport import simport
from oslo_config import cfg
from oslo_log import log

from monasca_persister import config
from monasca_persister.kafka import confluent_kafka_persister
from monasca_persister.kafka import legacy_kafka_persister


LOG = log.getLogger(__name__)

processors = []  # global list to facilitate clean signal handling
exiting = False


def clean_exit(signum, frame=None):
    """Exit all processes attempting to finish uncommitted active work before exit.
         Can be called on an os signal or no zookeeper losing connection.
    """
    global exiting
    if exiting:
        # Since this is set up as a handler for SIGCHLD when this kills one
        # child it gets another signal, the global exiting avoids this running
        # multiple times.
        LOG.debug('Exit in progress clean_exit received additional signal %s' % signum)
        return

    LOG.info('Received signal %s, beginning graceful shutdown.' % signum)
    exiting = True
    wait_for_exit = False

    for process in processors:
        try:
            if process.is_alive():
                # Sends sigterm which any processes after a notification is sent attempt to handle
                process.terminate()
                wait_for_exit = True
        except Exception:  # nosec
            # There is really nothing to do if the kill fails, so just go on.
            # The # nosec keeps bandit from reporting this as a security issue
            pass

    # wait for a couple seconds to give the subprocesses a chance to shut down correctly.
    if wait_for_exit:
        time.sleep(2)

    # Kill everything, that didn't already die
    for child in multiprocessing.active_children():
        LOG.debug('Killing pid %s' % child.pid)
        try:
            os.kill(child.pid, signal.SIGKILL)
        except Exception:  # nosec
            # There is really nothing to do if the kill fails, so just go on.
            # The # nosec keeps bandit from reporting this as a security issue
            pass

    if signum == signal.SIGTERM:
        sys.exit(0)

    sys.exit(signum)


def start_process(respository, kafka_config):
    LOG.info("start process: {}".format(respository))
    if kafka_config.legacy_kafka_client_enabled:
        m_persister = legacy_kafka_persister.LegacyKafkaPersister(
            kafka_config, cfg.CONF.zookeeper, respository)
    else:
        m_persister = confluent_kafka_persister.ConfluentKafkaPersister(
            kafka_config, respository)
    m_persister.run()


def prepare_processes(conf, repo_driver):
    if conf.num_processors > 0:
        repository = simport.load(repo_driver)
        for proc in range(0, conf.num_processors):
            processors.append(multiprocessing.Process(
                target=start_process, args=(repository, conf)))
    else:
        LOG.warning("Number of processors (num_processors) is {}".format(
            conf.num_processors))


def main():
    """Start persister."""

    config.parse_args()

    # Add processors for metrics topic
    if cfg.CONF.kafka_metrics.enabled:
        prepare_processes(cfg.CONF.kafka_metrics,
                          cfg.CONF.repositories.metrics_driver)
    # Add processors for alarm history topic
    if cfg.CONF.kafka_alarm_history.enabled:
        prepare_processes(cfg.CONF.kafka_alarm_history,
                          cfg.CONF.repositories.alarm_state_history_driver)
    # Add processors for events topic
    if cfg.CONF.kafka_events.enabled:
        prepare_processes(cfg.CONF.kafka_events,
                          cfg.CONF.repositories.events_driver)

    # Start
    try:
        LOG.info('''

               _____
              /     \   ____   ____ _____    ______ ____ _____
             /  \ /  \ /  _ \ /    \\\__  \  /  ___// ___\\\__  \\
            /    Y    (  <_> )   |  \/ __ \_\___ \\  \___ / __  \\_
            \____|__  /\____/|___|  (____  /____  >\___  >____  /
                    \/            \/     \/     \/     \/     \/
            __________                    .__          __
            \______   \ ___________  _____|__| _______/  |_  ___________
             |     ___// __ \_  __ \/  ___/  |/  ___/\   __\/ __ \_  __ \\
             |    |   \  ___/|  | \/\___ \|  |\___ \  |  | \  ___/|  | \/
             |____|    \___  >__|  /____  >__/____  > |__|  \___  >__|
                           \/           \/        \/            \/

        ''')
        for process in processors:
            process.start()

        # The signal handlers must be added after the processes start otherwise
        # they run on all processes
        signal.signal(signal.SIGCHLD, clean_exit)
        signal.signal(signal.SIGINT, clean_exit)
        signal.signal(signal.SIGTERM, clean_exit)

        while True:
            time.sleep(10)

    except Exception:
        LOG.exception('Error! Exiting.')
        clean_exit(signal.SIGKILL)

if __name__ == "__main__":
    sys.exit(main())



# (C) Copyright 2019 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Persister Recreate metric_id

   This tool is designed to 'fix' the rare instance when a metric_id
   has been removed from a row in Cassandra. That can cause issues
   when monasca-api retrieves the metric and tries to decode it.

   Configure this tool by copying the Monasca Persister settings from
   /opt/stack/service/monasca/etc/persister-config.yml in to a config
   .ini file (see template).

   Start the tool as stand-alone process by running
   'sudo -u mon_persister <venv python> \
   <venv path>/site-packages/monasca_persister/persister-recreate-metric-id.py \
   --config-file <config file>'

   When done, you may delete the config .ini file.

   Template for .ini file (suggested /opt/stack/service/monasca/etc/persister-recreate.ini)
   [DEFAULT]
   debug = False

   [repositories]
   metrics_driver = monasca_persister.repositories.cassandra.metrics_repository: \
   MetricCassandraRepository

   [cassandra]

   # Comma separated list of Cassandra node IP addresses (list value)
   contact_points = <single ip address for mgmt network on this node>

   # Cassandra port number (integer value)
   port = 9042

   # Keyspace name where metrics are stored (string value)
   #keyspace = monasca

   # Cassandra user name (string value)
   user = mon_persister

   # Cassandra password (string value)
   password = <password from persister-config.yml>

"""
import hashlib
import sys

from oslo_config import cfg
from oslo_log import log

from monasca_persister import config

from monasca_persister.repositories.cassandra import connection_util
from monasca_persister.repositories.cassandra.metrics_repository import METRICS_INSERT_CQL


LOG = log.getLogger(__name__)

METRIC_ALL_CQL = ('select region, tenant_id, metric_name, dimensions, '
                  'dimension_names, created_at, metric_id, updated_at '
                  'from metrics')


def usage():
    usage = """Monasca Persister Recreate metric_id Tool

    Used to recreate a metric_id, which in rare cases may be deleted
    from a row.  The metric_id is a hash of other fields, and thus can be
    recreated.  Note this tool is only for use with Cassandra storage installs.

    Please see the included README.rst for more details about creating an
    appropriate configuration file.

    persister-recreate-metric-id [-h] --config-file <ini>
        -h --help           Prints this
        --config-file <ini> (Required) Configuration file as described in README.rst

    Example

    To repair rows (execute as mon-persister user):
    persister-recreate-metric-id.py --config-file <path>/persister-recreate.ini

    """
    print(usage)


def main():
    """persister recreate metric_id tool."""

    config.parse_args()
    conf = cfg.CONF

    try:
        LOG.info('Starting check and repair of metric_id consistency.')

        # Connection setup
        # rocky style - note that we don't deliver pike style
        _cluster = connection_util.create_cluster()
        _session = connection_util.create_session(_cluster)
        _retention = conf.cassandra.retention_policy * 24 * 3600

        metric_all_stmt = _session.prepare(METRIC_ALL_CQL)
        metric_repair_stmt = _session.prepare(METRICS_INSERT_CQL)

        rows = _session.execute(metric_all_stmt)

        # if rows:
        #     LOG.info('First - {}'.format(rows[0]))
        #     # LOG.info('First name {} and id {}'.format(
        #     #     rows[0].metric_name, rows[0].metric_id)) # metric_id can't be logged raw

        # Bit of a misnomer - "null" is not in the cassandra db
        missing_value_rows = []
        for row in rows:
            if row.metric_id is None:
                LOG.info('Row with missing metric_id - {}'.format(row))
                missing_value_rows.append(row)

                # check created_at
                fixed_created_at = row.created_at
                if row.created_at is None and row.updated_at is not None:
                    LOG.info("Metric created_at was also None, repairing.")
                    fixed_created_at = row.updated_at

                # TODO(joadavis) update the updated_at timestamp to now

                # recreate metric id
                # copied from metrics_repository.py
                hash_string = '%s\0%s\0%s\0%s' % (row.region, row.tenant_id,
                                                  row.metric_name,
                                                  '\0'.join(row.dimensions))
                metric_id = hashlib.sha1(hash_string.encode('utf8')).hexdigest()
                id_bytes = bytearray.fromhex(metric_id)

                LOG.info("Recreated hash for metric id: {}".format(hash_string))
                # LOG.info("new id_bytes {}".format(id_bytes)) # can't unicode decode for logging

                # execute cql
                metric_repair_bound_stmt = metric_repair_stmt.bind((_retention,
                                                                    id_bytes,
                                                                    fixed_created_at,
                                                                    row.updated_at,
                                                                    row.region,
                                                                    row.tenant_id,
                                                                    row.metric_name,
                                                                    row.dimensions,
                                                                    row.dimension_names))

                _session.execute(metric_repair_bound_stmt)

        # LOG.info("of {} rows there are {} missing metric_id".format(len(rows), len(null_rows)))
        if len(missing_value_rows) > 0:
            LOG.warning("--> There were {} rows missing metric_id.".format(
                len(missing_value_rows)))
            LOG.warning("    Those rows have been updated.")
        else:
            LOG.info("No missing metric_ids were found, no changes made.")

        LOG.info('Done with metric_id consistency check and repair.')

        return 0

    except Exception:
        LOG.exception('Error! Exiting.')


if __name__ == "__main__":
    sys.exit(main())



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
from monasca_common.kafka import client_factory
import six

from monasca_persister.repositories import persister
from monasca_persister.repositories import singleton


@six.add_metaclass(singleton.Singleton)
class LegacyKafkaPersister(persister.Persister):

    def __init__(self, kafka_conf, zookeeper_conf, repository):
        super(LegacyKafkaPersister, self).__init__(kafka_conf, repository)
        self._consumer = client_factory.get_kafka_consumer(
            kafka_url=kafka_conf.uri,
            kafka_consumer_group=kafka_conf.group_id,
            kafka_topic=kafka_conf.topic,
            zookeeper_url=zookeeper_conf.uri,
            zookeeper_path=kafka_conf.zookeeper_path,
            use_legacy_client=True,
            repartition_callback=self._flush,
            commit_callback=self._flush,
            max_commit_interval=kafka_conf.max_wait_time_seconds
        )



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from monasca_persister.conf import types

elasticsearch_opts = [
    cfg.StrOpt(
        'index_name',
        help='Index prefix name where events are stored',
        default='events'),
    cfg.ListOpt(
        'hosts',
        help='List of Elasticsearch nodes in format host[:port]',
        default=['localhost:9200'],
        item_type=types.HostAddressPortType()),
    cfg.BoolOpt(
        'sniff_on_start',
        help='Flag indicating whether to obtain a list of nodes from the cluser at startup time',
        default=False),
    cfg.BoolOpt(
        'sniff_on_connection_fail',
        help='Flag controlling if connection failure triggers a sniff',
        default=False),
    cfg.IntOpt(
        'sniffer_timeout',
        help='Number of seconds between automatic sniffs',
        default=None),
    cfg.IntOpt(
        'max_retries',
        help='Maximum number of retries before an exception is propagated',
        default=3,
        min=1)]

elasticsearch_group = cfg.OptGroup(name='elasticsearch', title='elasticsearch')


def register_opts(conf):
    conf.register_group(elasticsearch_group)
    conf.register_opts(elasticsearch_opts, elasticsearch_group)


def list_opts():
    return elasticsearch_group, elasticsearch_opts



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log
import simplejson as json

from monasca_persister.repositories.influxdb import abstract_repository
from monasca_persister.repositories.influxdb import line_utils
from monasca_persister.repositories.utils import parse_alarm_state_hist_message

LOG = log.getLogger(__name__)


class AlarmStateHistInfluxdbRepository(
        abstract_repository.AbstractInfluxdbRepository):

    def __init__(self):

        super(AlarmStateHistInfluxdbRepository, self).__init__()

    def process_message(self, message):

        (alarm_id, metrics, new_state, old_state, link,
         lifecycle_state, state_change_reason,
         sub_alarms_json_snake_case, tenant_id,
         time_stamp) = parse_alarm_state_hist_message(
            message)

        name = 'alarm_state_history'
        fields = []
        fields.append('tenant_id=' + line_utils.escape_value(tenant_id))
        fields.append('alarm_id=' + line_utils.escape_value(alarm_id))
        fields.append('metrics=' + line_utils.escape_value(
            json.dumps(metrics, ensure_ascii=False)))
        fields.append('new_state=' + line_utils.escape_value(new_state))
        fields.append('old_state=' + line_utils.escape_value(old_state))
        fields.append('link=' + line_utils.escape_value(link))
        fields.append('lifecycle_state=' + line_utils.escape_value(
            lifecycle_state))
        fields.append('reason=' + line_utils.escape_value(
            state_change_reason))
        fields.append('reason_data=' + line_utils.escape_value("{}"))
        fields.append('sub_alarms=' + line_utils.escape_value(
            sub_alarms_json_snake_case))

        line = name + ',tenant_id=' + line_utils.escape_tag(tenant_id)
        line += ' ' + ','.join(fields)
        line += ' ' + str(int(time_stamp))

        LOG.debug(line)

        return line, tenant_id



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslotest import base

from monasca_persister.repositories.cassandra.token_range_query_manager \
    import TokenRangeQueryManager


class FakeException(Exception):
    pass


class TestTokenRangeQueryManager(base.BaseTestCase):

    def setUp(self):
        super(TestTokenRangeQueryManager, self).setUp()
        self._set_patchers()
        self._set_mocks()

        cql, result_handler = mock.Mock, mock.Mock()
        self.token_range_query_mgr = TokenRangeQueryManager(cql, result_handler, process_count=1)

    def _set_patchers(self):
        self.patcher_setup = mock.patch.object(TokenRangeQueryManager,
                                               '_setup',
                                               return_value=None)

    def _set_mocks(self):
        self.mock_setup = self.patcher_setup.start()

    def tearDown(self):
        super(TestTokenRangeQueryManager, self).tearDown()
        self.mock_setup.reset_mock()
        self.patcher_setup.stop()

    def test_close_pool(self):
        with mock.patch.object(self.token_range_query_mgr._pool, 'join',
                               side_effect=None):
            self.assertIsNone(self.token_range_query_mgr.close_pool())

    def test_query(self):
        with mock.patch.object(self.token_range_query_mgr._pool, 'map',
                               side_effect=FakeException):
            sample_element = mock.Mock()
            sample_element.value = 1
            token_ring = [sample_element, sample_element]
            self.assertRaises(FakeException, self.token_range_query_mgr.query, token_ring)



# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing

from oslo_config import cfg
from oslo_log import log

from monasca_persister.repositories.cassandra import connection_util


LOG = log.getLogger(__name__)

conf = cfg.CONF


class TokenRangeQueryManager(object):
    def __init__(self, cql, result_handler, process_count=None):
        if process_count:
            self._process_count = process_count
        else:
            self._process_count = multiprocessing.cpu_count()

        self._pool = multiprocessing.Pool(processes=self._process_count, initializer=self._setup,
                                          initargs=(cql, result_handler,))

    @classmethod
    def _setup(cls, cql, result_handler):
        cls.cluster = connection_util.create_cluster()
        cls.session = connection_util.create_session(cls.cluster)
        cls.prepared = cls.session.prepare(cql)
        cls.result_handler = result_handler

    def close_pool(self):
        self._pool.close()
        self._pool.join()

    def query(self, token_ring):

        range_size = len(token_ring) / self._process_count + 1
        start_index = 0
        params = []
        while start_index < len(token_ring):
            end_index = start_index + range_size - 1
            if end_index >= len(token_ring):
                end_index = len(token_ring) - 1
            params.append((token_ring[start_index].value, token_ring[end_index].value))
            start_index = end_index + 1

        self._pool.map(execute_query_token_range, params, 1)


def execute_query_token_range(token_range):
    results = TokenRangeQueryManager.session.execute(
        TokenRangeQueryManager.prepared.bind(token_range))
    TokenRangeQueryManager.result_handler(results)



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

influxdb_opts = [
    cfg.StrOpt('database_name',
               help='database name where metrics are stored',
               default='mon'),
    cfg.BoolOpt('db_per_tenant',
                help='Whether to use a separate database per tenant',
                default=False),
    cfg.IntOpt('default_retention_hours',
               help='Default retention period in hours for new '
                    'databases automatically created by the persister',
               default=0),
    cfg.IntOpt('batch_size',
               help='Maximum size of the batch to write to the database.',
               default=10000),
    cfg.HostAddressOpt('ip_address',
                       help='Valid IP address or hostname '
                            'to InfluxDB instance'),
    cfg.PortOpt('port',
                help='port to influxdb',
                default=8086),
    cfg.StrOpt('user',
               help='influxdb user ',
               default='mon_persister'),
    cfg.StrOpt('password',
               secret=True,
               help='influxdb password')]

influxdb_group = cfg.OptGroup(name='influxdb',
                              title='influxdb')


def register_opts(conf):
    conf.register_group(influxdb_group)
    conf.register_opts(influxdb_opts, influxdb_group)


def list_opts():
    return influxdb_group, influxdb_opts



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Copyright 2017 FUJITSU LIMITED
# (C) Copyright 2017 SUSE LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_config import types

cassandra_opts = [
    cfg.ListOpt('contact_points',
                help='Comma separated list of Cassandra node IP addresses',
                default=['127.0.0.1'],
                item_type=types.HostAddress()),
    cfg.IntOpt('port',
               help='Cassandra port number',
               default=8086),
    cfg.StrOpt('keyspace',
               help='Keyspace name where metrics are stored',
               default='monasca'),
    cfg.StrOpt('user',
               help='Cassandra user name',
               default=''),
    cfg.StrOpt('password',
               help='Cassandra password',
               secret=True,
               default=''),
    cfg.IntOpt('connection_timeout',
               help='Cassandra timeout in seconds when creating a new connection',
               default=5),
    cfg.IntOpt('read_timeout',
               help='Cassandra read timeout in seconds',
               default=60),
    cfg.IntOpt('max_write_retries',
               help='Maximum number of retries in write ops',
               default=1),
    cfg.IntOpt('max_definition_cache_size',
               help='Maximum number of cached metric definition entries in memory',
               default=20000000),
    cfg.IntOpt('retention_policy',
               help='Data retention period in days',
               default=45),
    cfg.StrOpt('consistency_level',
               help='Cassandra default consistency level',
               default='ONE'),
    cfg.StrOpt('local_data_center',
               help='Cassandra local data center name'),
    cfg.IntOpt('max_batches',
               help='Maximum batch size in Cassandra',
               default=250),
]

cassandra_group = cfg.OptGroup(name='cassandra')


def register_opts(conf):
    conf.register_group(cassandra_group)
    conf.register_opts(cassandra_opts, cassandra_group)


def list_opts():
    return cassandra_group, cassandra_opts



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.


class DataPointsAsList(list):

    def append(self, key, value):
        super(DataPointsAsList, self).append(value)

    def get_count(self):
        return len(self)

    def clear(self):
        del self[:]


class DataPointsAsDict(dict):

    def __init__(self):
        self.counter = 0

    def __setitem__(self, key, value):
        raise NotImplementedError('Use append(key, value) instead.')

    def __delitem__(self, key):
        raise NotImplementedError('Use clear() instead.')

    def pop(self):
        raise NotImplementedError('Use clear() instead.')

    def popitem(self):
        raise NotImplementedError('Use clear() instead.')

    def update(self):
        raise NotImplementedError('Use clear() instead.')

    def append(self, key, value):
        super(DataPointsAsDict, self).setdefault(key, []).append(value)
        self.counter += 1

    def clear(self):
        super(DataPointsAsDict, self).clear()
        self.counter = 0

    def get_count(self):
        return self.counter



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_config import cfg
from oslo_log import log

from monasca_persister import conf
from monasca_persister import version

CONF = conf.CONF
LOG = log.getLogger(__name__)

_CONF_LOADED = False


def _get_config_files():
    """Get the possible configuration files accepted by oslo.config

    This also includes the deprecated ones
    """
    # default files
    conf_files = cfg.find_config_files(project='monasca',
                                       prog='monasca-persister')
    # deprecated config files (only used if standard config files are not there)
    if len(conf_files) == 0:
        old_conf_files = cfg.find_config_files(project='monasca',
                                               prog='persister')
        if len(old_conf_files) > 0:
            LOG.warning('Found deprecated old location "{}" '
                        'of main configuration file'.format(old_conf_files))
            conf_files += old_conf_files
    return conf_files


def parse_args(description='Persists metrics & alarm history in TSDB'):
    global _CONF_LOADED
    if _CONF_LOADED:
        LOG.debug('Configuration has been already loaded')
        return

    log.set_defaults()
    log.register_options(CONF)

    CONF(prog=sys.argv[1:],
         project='monasca',
         version=version.version_str,
         default_config_files=_get_config_files(),
         description=description)

    log.setup(CONF,
              product_name='monasca-persister',
              version=version.version_str)

    conf.register_opts()

    _CONF_LOADED = True



# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cassandra.policies import RetryPolicy


class MonascaRetryPolicy(RetryPolicy):

    def __init__(self, read_attempts, write_attempts, unavailable_attempts):

        super(MonascaRetryPolicy, self).__init__()

        self.read_attempts = read_attempts
        self.write_attempts = write_attempts
        self.unavailable_attempts = unavailable_attempts

    def on_read_timeout(self, query, consistency, required_responses,
                        received_responses, data_retrieved, retry_num):

        if retry_num >= self.read_attempts:
            return self.RETHROW, None
        elif received_responses >= required_responses and not data_retrieved:
            return self.RETRY, consistency
        else:
            return self.RETHROW, None

    def on_write_timeout(self, query, consistency, write_type,
                         required_responses, received_responses, retry_num):

        if retry_num >= self.write_attempts:
            return self.RETHROW, None
        else:
            return self.RETRY, consistency

    def on_unavailable(self, query, consistency, required_replicas, alive_replicas, retry_num):

        return (
            self.RETRY_NEXT_HOST,
            consistency) if retry_num < self.unavailable_attempts else (
            self.RETHROW,
            None)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement
from cachetools import LRUCache
from collections import namedtuple
import hashlib
import threading

from cassandra.concurrent import execute_concurrent
from oslo_log import log
import simplejson as json

from monasca_persister.repositories.cassandra import abstract_repository
from monasca_persister.repositories.cassandra import token_range_query_manager
from monasca_persister.repositories.cassandra.metric_batch import MetricBatch
from monasca_persister.repositories.utils import parse_measurement_message

LOG = log.getLogger(__name__)

MEASUREMENT_INSERT_CQL = (
    'update monasca.measurements USING TTL ? '
    'set value = ?, value_meta = ?, region = ?, tenant_id = ?, metric_name = ?, dimensions = ? '
    'where metric_id = ? and time_stamp = ?')

MEASUREMENT_UPDATE_CQL = ('update monasca.measurements USING TTL ? '
                          'set value = ?, value_meta = ? where metric_id = ? and time_stamp = ?')

METRICS_INSERT_CQL = ('update monasca.metrics USING TTL ? '
                      'set metric_id = ?, created_at = ?, updated_at = ? '
                      'where region = ? and tenant_id = ? and metric_name = ? and dimensions = ? '
                      'and dimension_names = ?')

METRICS_UPDATE_CQL = ('update monasca.metrics USING TTL ? '
                      'set metric_id = ?, updated_at = ? '
                      'where region = ? and tenant_id = ? and metric_name = ? and dimensions = ? '
                      'and dimension_names = ?')

DIMENSION_INSERT_CQL = ('insert into  monasca.dimensions '
                        '(region, tenant_id, name, value) values (?, ?, ?, ?)')

DIMENSION_METRIC_INSERT_CQL = ('insert into monasca.dimensions_metrics '
                               '(region, tenant_id, dimension_name, dimension_value, metric_name) '
                               'values (?, ?, ?, ?, ?)')

METRIC_DIMENSION_INSERT_CQL = ('insert into monasca.metrics_dimensions '
                               '(region, tenant_id, metric_name, dimension_name, dimension_value) '
                               'values (?, ?, ?, ?, ?)')

RETRIEVE_DIMENSION_CQL = 'select region, tenant_id, name, value from dimensions'

RETRIEVE_METRIC_DIMENSION_CQL = ('select region, tenant_id, metric_name, '
                                 'dimension_name, dimension_value from metrics_dimensions '
                                 'WHERE token(region, tenant_id, metric_name) > ? '
                                 'and token(region, tenant_id, metric_name) <= ? ')

Metric = namedtuple('Metric',
                    ['id',
                     'region',
                     'tenant_id',
                     'name',
                     'dimension_list',
                     'dimension_names',
                     'time_stamp',
                     'value',
                     'value_meta'])


class MetricCassandraRepository(abstract_repository.AbstractCassandraRepository):
    def __init__(self):
        super(MetricCassandraRepository, self).__init__()

        self._lock = threading.RLock()

        LOG.debug("prepare cql statements...")

        self._measurement_insert_stmt = self._session.prepare(MEASUREMENT_INSERT_CQL)
        self._measurement_insert_stmt.is_idempotent = True

        self._measurement_update_stmt = self._session.prepare(MEASUREMENT_UPDATE_CQL)
        self._measurement_update_stmt.is_idempotent = True

        self._metric_insert_stmt = self._session.prepare(METRICS_INSERT_CQL)
        self._metric_insert_stmt.is_idempotent = True

        self._metric_update_stmt = self._session.prepare(METRICS_UPDATE_CQL)
        self._metric_update_stmt.is_idempotent = True

        self._dimension_stmt = self._session.prepare(DIMENSION_INSERT_CQL)
        self._dimension_stmt.is_idempotent = True

        self._dimension_metric_stmt = self._session.prepare(DIMENSION_METRIC_INSERT_CQL)
        self._dimension_metric_stmt.is_idempotent = True

        self._metric_dimension_stmt = self._session.prepare(METRIC_DIMENSION_INSERT_CQL)
        self._metric_dimension_stmt.is_idempotent = True

        self._retrieve_metric_dimension_stmt = self._session.prepare(RETRIEVE_METRIC_DIMENSION_CQL)

        self._metric_batch = MetricBatch(
            self._cluster.metadata,
            self._cluster.load_balancing_policy,
            self._max_batches)

        self._metric_id_cache = LRUCache(self._cache_size)
        self._dimension_cache = LRUCache(self._cache_size)
        self._metric_dimension_cache = LRUCache(self._cache_size)

        self._load_dimension_cache()
        self._load_metric_dimension_cache()

    def process_message(self, message):
        (dimensions, metric_name, region, tenant_id, time_stamp, value,
         value_meta) = parse_measurement_message(message)

        with self._lock:
            dim_names = []
            dim_list = []
            for name in sorted(dimensions.iterkeys()):
                dim_list.append('%s\t%s' % (name, dimensions[name]))
                dim_names.append(name)

            hash_string = '%s\0%s\0%s\0%s' % (region, tenant_id, metric_name, '\0'.join(dim_list))
            metric_id = hashlib.sha1(hash_string.encode('utf8')).hexdigest()

            # TODO(brtknr): If database per tenant becomes the default and the
            # only option, recording tenant_id will be redundant.
            metric = Metric(id=metric_id,
                            region=region,
                            tenant_id=tenant_id,
                            name=metric_name,
                            dimension_list=dim_list,
                            dimension_names=dim_names,
                            time_stamp=time_stamp,
                            value=value,
                            value_meta=json.dumps(value_meta, ensure_ascii=False))

            id_bytes = bytearray.fromhex(metric.id)
            if self._metric_id_cache.get(metric.id, None):
                measurement_bound_stmt = self._measurement_update_stmt.bind((self._retention,
                                                                             metric.value,
                                                                             metric.value_meta,
                                                                             id_bytes,
                                                                             metric.time_stamp))
                self._metric_batch.add_measurement_query(measurement_bound_stmt)

                metric_update_bound_stmt = self._metric_update_stmt.bind((self._retention,
                                                                          id_bytes,
                                                                          metric.time_stamp,
                                                                          metric.region,
                                                                          metric.tenant_id,
                                                                          metric.name,
                                                                          metric.dimension_list,
                                                                          metric.dimension_names))
                self._metric_batch.add_metric_query(metric_update_bound_stmt)

                return metric, tenant_id

            self._metric_id_cache[metric.id] = metric.id

            metric_insert_bound_stmt = self._metric_insert_stmt.bind((self._retention,
                                                                      id_bytes,
                                                                      metric.time_stamp,
                                                                      metric.time_stamp,
                                                                      metric.region,
                                                                      metric.tenant_id,
                                                                      metric.name,
                                                                      metric.dimension_list,
                                                                      metric.dimension_names))
            self._metric_batch.add_metric_query(metric_insert_bound_stmt)

            for dim in metric.dimension_list:
                (name, value) = dim.split('\t')
                dim_key = self._get_dimnesion_key(metric.region, metric.tenant_id, name, value)
                if not self._dimension_cache.get(dim_key, None):
                    dimension_bound_stmt = self._dimension_stmt.bind((metric.region,
                                                                      metric.tenant_id,
                                                                      name,
                                                                      value))
                    self._metric_batch.add_dimension_query(dimension_bound_stmt)
                    self._dimension_cache[dim_key] = dim_key

                metric_dim_key = self._get_metric_dimnesion_key(
                    metric.region, metric.tenant_id, metric.name, name, value)
                if not self._metric_dimension_cache.get(metric_dim_key, None):
                    dimension_metric_bound_stmt = self._dimension_metric_stmt.bind(
                        (metric.region, metric.tenant_id, name, value, metric.name))
                    self._metric_batch.add_dimension_metric_query(dimension_metric_bound_stmt)

                    metric_dimension_bound_stmt = self._metric_dimension_stmt.bind(
                        (metric.region, metric.tenant_id, metric.name, name, value))
                    self._metric_batch.add_metric_dimension_query(metric_dimension_bound_stmt)

                    self._metric_dimension_cache[metric_dim_key] = metric_dim_key

            measurement_insert_bound_stmt = self._measurement_insert_stmt.bind(
                (self._retention,
                 metric.value,
                 metric.value_meta,
                 metric.region,
                 metric.tenant_id,
                 metric.name,
                 metric.dimension_list,
                 id_bytes,
                 metric.time_stamp))
            self._metric_batch.add_measurement_query(measurement_insert_bound_stmt)

            return metric, tenant_id

    def write_batch(self, metrics):
        with self._lock:
            batch_list = self._metric_batch.get_all_batches()
            results = execute_concurrent(self._session, batch_list, raise_on_first_error=True)
            self._handle_results(results)
            self._metric_batch.clear()
            LOG.info("flushed %s metrics", len(metrics))

    @staticmethod
    def _handle_results(results):
        for (success, result) in results:
            if not success:
                raise result

    def _load_dimension_cache(self):

        rows = self._session.execute(RETRIEVE_DIMENSION_CQL)

        if not rows:
            return

        for row in rows:
            key = self._get_dimnesion_key(row.region, row.tenant_id, row.name, row.value)
            self._dimension_cache[key] = key

        LOG.info(
            "loaded %s dimension entries cache from database into cache." %
            self._dimension_cache.currsize)

    @staticmethod
    def _get_dimnesion_key(region, tenant_id, name, value):
        return '%s\0%s\0%s\0%s' % (region, tenant_id, name, value)

    def _load_metric_dimension_cache(self):
        qm = token_range_query_manager.TokenRangeQueryManager(RETRIEVE_METRIC_DIMENSION_CQL,
                                                              self._process_metric_dimension_query)

        token_ring = self._cluster.metadata.token_map.ring

        qm.query(token_ring)

    def _process_metric_dimension_query(self, rows):

        cnt = 0
        for row in rows:
            key = self._get_metric_dimnesion_key(
                row.region,
                row.tenant_id,
                row.metric_name,
                row.dimension_name,
                row.dimension_value)
            self._metric_dimension_cache[key] = key
            cnt += 1

        LOG.info("loaded %s metric dimension entries from database into cache." % cnt)
        LOG.info(
            "total loaded %s metric dimension entries in cache." %
            self._metric_dimension_cache.currsize)

    @staticmethod
    def _get_metric_dimnesion_key(region, tenant_id, metric_name, dimension_name, dimension_value):

        return '%s\0%s\0%s\0%s\0%s' % (region, tenant_id, metric_name,
                                       dimension_name, dimension_value)



# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
from cassandra.cluster import ConsistencyLevel
from cassandra.cluster import DCAwareRoundRobinPolicy
from cassandra.cluster import TokenAwarePolicy
from oslo_config import cfg

from monasca_persister.repositories.cassandra.retry_policy import MonascaRetryPolicy

conf = cfg.CONF

def create_cluster():
    user = conf.cassandra.user
    if user:
        auth_provider = PlainTextAuthProvider(username=user, password=conf.cassandra.password)
    else:
        auth_provider = None

    cluster = Cluster(conf.cassandra.contact_points,
                      port=conf.cassandra.port,
                      auth_provider=auth_provider,
                      connect_timeout=conf.cassandra.connection_timeout,
                      load_balancing_policy=TokenAwarePolicy(
                          DCAwareRoundRobinPolicy(local_dc=conf.cassandra.local_data_center)),
                      )
    cluster.default_retry_policy = MonascaRetryPolicy(1, conf.cassandra.max_write_retries,
                                                      conf.cassandra.max_write_retries)
    return cluster


def create_session(cluster):
    session = cluster.connect(conf.cassandra.keyspace)
    session.default_timeout = conf.cassandra.read_timeout
    session.default_consistency_level = \
        ConsistencyLevel.name_to_value[conf.cassandra.consistency_level]
    return session



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from elasticsearch import Elasticsearch
from oslo_config import cfg
from oslo_log import log
import simplejson as json

from monasca_persister.repositories import abstract_repository
from monasca_persister.repositories import data_points
from monasca_persister.repositories import utils

LOG = log.getLogger(__name__)


class ElasticSearchEventsRepository(abstract_repository.AbstractRepository):
    def __init__(self):
        super(ElasticSearchEventsRepository, self).__init__()
        self.conf = cfg.CONF.elasticsearch
        self.es = Elasticsearch(
            hosts=self.conf.hosts,
            sniff_on_start=self.conf.sniff_on_start,
            sniff_on_connection_fail=self.conf.sniff_on_connection_fail,
            sniffer_timeout=self.conf.sniffer_timeout,
            max_retries=self.conf.max_retries
        )
        self.data_points_class = data_points.DataPointsAsList

    def process_message(self, message):
        return utils.parse_events_message(message)

    def write_batch(self, data_points):
        for data_point in data_points:
            (project_id, timestamp, event_type, payload, dimensions) = data_point

            index = '%s-%s-%s' % (self.conf.index_name, project_id,
                                  ElasticSearchEventsRepository._normalize_timestamp(timestamp))

            body = {
                'project_id': project_id,
                '@timestamp': timestamp,
                'event_type': event_type,
                'payload': payload,
                'dimensions': dimensions
            }

            self.es.create(
                index=index,
                doc_type='event',
                body=json.dumps(body)
            )

    @staticmethod
    def _normalize_timestamp(timestamp):
        d = None
        if timestamp and len(timestamp) >= 10:
            try:
                d = datetime.strptime(timestamp[0:10], '%Y-%m-%d')
            except ValueError as e:
                LOG.warning("Unable to parse timestamp '%s' - %s" % (timestamp, str(e)))
        if not d:
            d = datetime.today()
        return d.strftime('%Y-%m-%d')



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import influxdb
from oslo_config import cfg
import six

from monasca_persister.repositories import abstract_repository
from monasca_persister.repositories import data_points

DATABASE_NOT_FOUND_MSG = "database not found"


@six.add_metaclass(abc.ABCMeta)
class AbstractInfluxdbRepository(abstract_repository.AbstractRepository):

    def __init__(self):
        super(AbstractInfluxdbRepository, self).__init__()
        self.conf = cfg.CONF
        self._influxdb_client = influxdb.InfluxDBClient(
            self.conf.influxdb.ip_address,
            self.conf.influxdb.port,
            self.conf.influxdb.user,
            self.conf.influxdb.password)
        if self.conf.influxdb.db_per_tenant:
            self.data_points_class = data_points.DataPointsAsDict
        else:
            self.data_points_class = data_points.DataPointsAsList

    def write_batch(self, data_points):
        if self.conf.influxdb.db_per_tenant:
            for tenant_id, tenant_data_points in data_points.items():
                database = '%s_%s' % (self.conf.influxdb.database_name,
                                      tenant_id)
                self._write_batch(tenant_data_points, database)
        else:
            self._write_batch(data_points, self.conf.influxdb.database_name)

    def _write_batch(self, data_points, database):
        # NOTE (brtknr): Loop twice to ensure database is created if missing.
        for retry in range(2):
            try:
                batch_size = self.conf.influxdb.batch_size
                self._influxdb_client.write_points(data_points, 'ms',
                                                   protocol='line',
                                                   database=database,
                                                   batch_size=batch_size)
                break
            except influxdb.exceptions.InfluxDBClientError as ex:
                # When a databse is not found, the returned exception resolves
                # to: {"error":"database not found: \"test\""}
                if DATABASE_NOT_FOUND_MSG in str(ex):
                    self._influxdb_client.create_database(database)
                    # NOTE (brtknr): Only apply default retention policy at
                    # database creation time so that existing policies are
                    # not overridden since administrators may want different
                    # retention policy per tenant.
                    hours = self.conf.influxdb.default_retention_hours
                    if hours > 0:
                        rp = '{}h'.format(hours)
                        default_rp = dict(database=database, default=True,
                                          name=rp, duration=rp,
                                          replication='1')
                        self._influxdb_client.create_retention_policy(**default_rp)
                else:
                    raise



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from oslo_config import cfg
import six

from monasca_persister.repositories import abstract_repository
from monasca_persister.repositories.cassandra import connection_util
from monasca_persister.repositories import data_points

conf = cfg.CONF


@six.add_metaclass(abc.ABCMeta)
class AbstractCassandraRepository(abstract_repository.AbstractRepository):
    def __init__(self):
        super(AbstractCassandraRepository, self).__init__()

        self._cluster = connection_util.create_cluster()
        self._session = connection_util.create_session(self._cluster)
        self._retention = conf.cassandra.retention_policy * 24 * 3600
        self._cache_size = conf.cassandra.max_definition_cache_size
        self._max_batches = conf.cassandra.max_batches
        self.data_points_class = data_points.DataPointsAsList



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
import os

from oslo_config import cfg
from oslo_log import log
import six

from monasca_persister.repositories import singleton

LOG = log.getLogger(__name__)


@six.add_metaclass(singleton.Singleton)
class Persister(six.with_metaclass(ABCMeta, object)):

    def __init__(self, kafka_conf, repository):

        self._kafka_topic = kafka_conf.topic
        self._batch_size = kafka_conf.batch_size
        self.repository = repository()
        self._data_points = self.repository.data_points_class()

    def _flush(self):
        if not self._data_points:
            return

        try:
            self.repository.write_batch(self._data_points)

            LOG.info("Processed {} messages from topic '{}'".format(
                self._data_points.get_count(), self._kafka_topic))

            self._data_points.clear()
            self._consumer.commit()
        except Exception as ex:
            if "partial write: points beyond retention policy dropped" in str(ex):
                LOG.warning("Some points older than retention policy were dropped")
                self._data_points.clear()
                self._consumer.commit()

            elif cfg.CONF.repositories.ignore_parse_point_error \
                    and "unable to parse" in str(ex):
                LOG.warning("Some points were unable to be parsed and were dropped")
                self._data_points.clear()
                self._consumer.commit()

            else:
                LOG.exception("Error writing to database: {}"
                              .format(self._data_points))
                raise ex

    def run(self):
        try:
            for message in self._consumer:
                try:
                    data_point, tenant_id = self.repository.process_message(message)
                    self._data_points.append(tenant_id, data_point)
                except Exception:
                    LOG.exception('Error processing message. Message is '
                                  'being dropped. {}'.format(message))

                if self._data_points.get_count() >= self._batch_size:
                    self._flush()
        except Exception:
            LOG.exception(
                'Persister encountered fatal exception processing '
                'messages. '
                'Shutting down all threads and exiting')
            os._exit(1)



# (C) Copyright 2019 StackHPC Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Monasca InfluxDB tool for migration to database per tenant

    Used to move data from monolithic database e.g. `monasca` to a database per
    tenant model, e.g. `monasca_<tenant_id>`.

    Please see the included README.rst for more details about creating an
    appropriate configuration file.
"""

import os
import re
import sys

from monasca_persister import config
from monasca_persister.repositories.influxdb import metrics_repository

from oslo_config import cfg

from oslo_log import log

LOG = log.getLogger(__name__)

MIGRATE_QUERY = ('SELECT * INTO "{target_db}"..:MEASUREMENT'
                 ' FROM "{measurement}"'
                 ' WHERE _tenant_id=\'{tenant_id}\''
                 ' AND time > {lower_time_offset}'
                 ' AND time <= {upper_time_offset}'
                 ' GROUP BY *')

class MigrationHelper(object):

    def __init__(self):
        repo = metrics_repository.MetricInfluxdbRepository()
        self.conf = repo.conf
        self.client = repo._influxdb_client
        self.client.switch_database(self.conf.influxdb.database_name)

    def _migrate(self, measurement, tenant_id, start_time_offset,
                 end_time_offset, retention_policy={}, time_unit='w',
                 db_per_tenant=True, **kwargs):

        total_written = 0
        first_upper_time_offset = None
        last_lower_time_offset = None
        time_offset = start_time_offset

        if db_per_tenant:
            target_db = "{}_{}".format(self.conf.influxdb.database_name, tenant_id)
        self.client.create_database(target_db)
        if retention_policy:
            self.client.create_retention_policy(database=target_db, **retention_policy)
        LOG.info('         into {}:'.format(target_db))

        while end_time_offset > 0 and time_offset < end_time_offset:
            lower_time_offset = 'now()-{}{}'.format(time_offset + 1, time_unit)
            upper_time_offset = 'now()-{}{}'.format(time_offset, time_unit)
            if not first_upper_time_offset:
                first_upper_time_offset = upper_time_offset
            migrate_query = MIGRATE_QUERY.format(
                target_db=target_db,
                measurement=measurement,
                tenant_id=tenant_id,
                lower_time_offset=lower_time_offset,
                upper_time_offset=upper_time_offset,
            )
            LOG.debug(migrate_query)

            written = next(self.client.query(migrate_query).get_points('result')).get('written')
            total_written += written
            time_offset += 1
            if written > 0:
                last_lower_time_offset = lower_time_offset
                LOG.info("         migrated {} entries from {} -> {} (cumulative {})".format(
                    written,
                    lower_time_offset,
                    upper_time_offset,
                    total_written,
                ))
        LOG.info("         finished migrating a total of {} entries from {} -> {}.".format(
            total_written,
            last_lower_time_offset,
            first_upper_time_offset,
        ))

    def get_measurements(self, fname):
        measurements = []
        if fname:
            with open(fname, 'a+') as f:
                measurements = [l.strip() for l in f.readlines()]
        if not measurements:
            result = self.client.query('SHOW MEASUREMENTS').get_points('measurements')
            measurements = [m.get('name') for m in result]
            if fname:
                with open(fname, 'w') as f:
                    for r in measurements:
                        f.write(r + '\n')
        return measurements

    def get_tenancies(self, measurements):
        result = self.client.query("SHOW TAG VALUES WITH KEY = _tenant_id")
        return {m: [t.get('value') for t in result.get_points(m)] for m in measurements}

    def get_complete(self, fname):
        if fname:
            with open(fname, 'a+') as fd:
                return {l.strip() for l in fd.readlines()}
        else:
            return {}

    def migrate(self,
                tenant_defaults={},
                default_start_time_offset=0,  # Default: now
                default_end_time_offset=520,  # Default: 10 years
                skip_regex=[],
                measurements_file=None, success_file=None, failure_file=None, **kwargs):
        measurements = self.get_measurements(measurements_file)
        tenancy = self.get_tenancies(measurements)
        done = self.get_complete(success_file)
        default_rp = {}
        hours = self.conf.influxdb.default_retention_hours
        if hours > 0:
            rp = '{}h'.format(hours)
            default_rp = dict(name=rp, duration=rp, replication='1', default=True)
        skip = set()
        fail = set()
        if failure_file:
            if os.path.exists(failure_file):
                os.remove(failure_file)

        filtered_measurements = []
        for measurement in measurements:
            if any([f.match(measurement) for f in skip_regex]):
                skip.add(measurement)
                LOG.debug('Skipping {} because it matches a skip regex.'.format(measurement))
                continue
            elif measurement in done:
                LOG.debug('Skipping {} because its already done.'.format(measurement))
                continue
            else:
                filtered_measurements.append(measurement)

        for i, measurement in enumerate(filtered_measurements):
            LOG.info('Migrating {}'.format(measurement))
            try:
                for tenant_id in tenancy.get(measurement):
                    start_time_offset = tenant_defaults.get(
                        tenant_id, {}).get('start_time_offset_override',
                                           default_start_time_offset)
                    end_time_offset = tenant_defaults.get(
                        tenant_id, {}).get('end_time_offset_override',
                                           default_end_time_offset)
                    # NOTE (brtknr): Ensure that the default upper and lower
                    # time offsets are respected during migration by the
                    # projects with custom retention policies.
                    start_time_offset = max(default_start_time_offset,
                                            start_time_offset)
                    end_time_offset = min(default_end_time_offset,
                                          end_time_offset)
                    retention_policy = tenant_defaults.get(
                        tenant_id, {}).get('rp', default_rp)
                    self._migrate(measurement, tenant_id,
                                  start_time_offset=start_time_offset,
                                  end_time_offset=end_time_offset,
                                  retention_policy=retention_policy, **kwargs)
                if success_file:
                    with open(success_file, 'a+') as fd:
                        fd.write('{}\n'.format(measurement))
                done.add(measurement)
            except Exception as e:
                LOG.error(e)
                if failure_file:
                    with open(failure_file, 'a+') as fe:
                        fe.write('{}\t{}\n'.format(measurement, e))
                fail.add(measurement)
            LOG.info("{}/{} (done {} + skip {} + fail {})/{}".format(
                i + 1, len(filtered_measurements), len(done), len(skip),
                len(fail), len(measurements)))


def main():
    CONF = cfg.CONF
    cli_opts = [
        cfg.StrOpt('migrate-time-unit', choices=['h', 'd', 'w'], default='w',
                   help='Unit of time, h=hour, d=day, w=week (default: "w").'),
        cfg.IntOpt('migrate-start-time-offset', default=0,
                   help='Start time offset in the given unit of time (default: 0).'),
        cfg.IntOpt('migrate-end-time-offset', default=520,
                   help='End time offset in the given unit of time (default: 520).'),
        cfg.DictOpt('migrate-retention-policy', default={},
                    help=('Custom retention policy for projects in the provided'
                          'time unit. (e.g. project-id-x:2,project-id-y:4)')),
        cfg.ListOpt('migrate-skip-regex', default=[],
                    help=('Skip metrics that match this comma separated list of regex patterns.'
                          '(e.g. ^log\\\\..+,^cpu\\\\..+ to skip metrics beginning with log.)')),
    ]
    CONF.register_cli_opts(cli_opts)
    config.parse_args("Monasca InfluxDB database per tenant migration tool")

    # Configure custom retention policy for your existing projects. For
    # example, rp2w is a retention policy of two weeks which we can assign to
    # project example-project-id.
    tenant_defaults = dict()
    for k, v in CONF.migrate_retention_policy.items():
        if v.isdigit():
            rp = '{}{}'.format(v, CONF.migrate_time_unit)
            tenant_defaults[k] = dict(
                end_time_offset_override=int(v),
                rp=dict(name=rp, duration=rp, replication='1', default=True),
            )
            LOG.info('Project {} will be applied retention policy: {}.'.format(k, rp))
        else:
            raise ValueError('Retention policy for project {} must be an'
                             'integer of given time unit. Current value:'
                             '{}.'.format(k, v))

    skip_regex = []
    for p in CONF.migrate_skip_regex:
        skip_regex.append(re.compile(str(p)))
        LOG.info('Metrics matching pattern "{}" will be skipped.'.format(p))

    helper = MigrationHelper()
    helper.migrate(skip_regex=skip_regex,
                   tenant_defaults=tenant_defaults,
                   default_end_time_offset=CONF.migrate_end_time_offset,
                   default_start_time_offset=CONF.migrate_start_time_offset,
                   time_unit=CONF.migrate_time_unit,
                   measurements_file='migrate-measurements',
                   success_file='migrate-success',
                   failure_file='migrate-failure')
    return 0


if __name__ == "__main__":
    sys.exit(main())



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslotest import base

from monasca_persister.conf import types


class TestHostAddressPortType(base.BaseTestCase):

    def setUp(self):
        super(TestHostAddressPortType, self).setUp()
        self.types = types.HostAddressPortType()

    def test_ip_address(self):

        self.assertEqual('127.0.0.1:2121', self.types('127.0.0.1:2121'))

    def test_hostname(self):
        self.assertEqual('localhost:2121', self.types('localhost:2121'))

    # failure scenario
    def test_missing_port(self):
        self.assertRaises(ValueError, self.types, '127.0.0.1')

    def test_missing_address(self):
        self.assertRaises(ValueError, self.types, ':123')

    def test_incorrect_ip(self):
        self.assertRaises(ValueError, self.types, '127.surprise.0.1:2121')

    def test_incorrect_port(self):
        self.assertRaises(ValueError, self.types, '127.0.0.1:65536')
        self.assertRaises(ValueError, self.types, '127.0.0.1:sample')



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from monasca_persister.conf import types

zookeeper_opts = [
    cfg.ListOpt('uri',
                help='Comma separated list of zookeper instance host:port',
                default=['127.0.0.1:2181'],
                item_type=types.HostAddressPortType()),
    cfg.IntOpt('partition_interval_recheck_seconds',
               help='Time between rechecking if partition is available',
               default=15)]

zookeeper_group = cfg.OptGroup(name='zookeeper', title='zookeeper')


def register_opts(conf):
    conf.register_group(zookeeper_group)
    conf.register_opts(zookeeper_opts, zookeeper_group)


def list_opts():
    return zookeeper_group, zookeeper_opts



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_config import types


class HostAddressPortOpt(cfg.Opt):
    """Option for HostAddressPortType.

    Accept hostname or ip address with TCP/IP port number.
    """
    def __init__(self, name, **kwargs):
        ip_port_type = HostAddressPortType()
        super(HostAddressPortOpt, self).__init__(name,
                                                 type=ip_port_type,
                                                 **kwargs)


class HostAddressPortType(types.HostAddress):
    """HostAddress with additional port."""

    def __init__(self, version=None):
        type_name = 'ip and port value'
        super(HostAddressPortType, self).__init__(version, type_name=type_name)

    def __call__(self, value):
        addr, port = value.split(':')
        addr = self.validate_addr(addr)
        port = self._validate_port(port)
        if not addr and not port:
            raise ValueError('%s is not valid ip with optional port')
        return '%s:%d' % (addr, port)

    @staticmethod
    def _validate_port(port):
        return types.Port()(port)

    def validate_addr(self, addr):
        try:
            addr = self.ip_address(addr)
        except ValueError:
            try:
                addr = self.hostname(addr)
            except ValueError:
                raise ValueError("%s is not a valid host address", addr)
        return addr



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log
import simplejson as json

from monasca_persister.repositories.influxdb import abstract_repository
from monasca_persister.repositories.influxdb import line_utils
from monasca_persister.repositories.utils import parse_measurement_message

LOG = log.getLogger(__name__)


class MetricInfluxdbRepository(abstract_repository.AbstractInfluxdbRepository):

    def __init__(self):

        super(MetricInfluxdbRepository, self).__init__()

    @staticmethod
    def process_message(message):

        (dimensions, metric_name, region, tenant_id, time_stamp, value,
         value_meta) = parse_measurement_message(message)

        tags = dimensions

        # TODO(brtknr): If database per tenant becomes the default and the only
        # option, recording tenant_id will be redundant.
        tags[u'_tenant_id'] = tenant_id
        tags[u'_region'] = region

        if not value_meta:
            value_meta_str = u'"{}"'
        else:
            value_meta_str = line_utils.escape_value(json.dumps(value_meta, ensure_ascii=False))

        key_values = [line_utils.escape_tag(metric_name)]

        # tags should be sorted client-side to take load off server
        for key in sorted(tags.keys()):
            key_tag = line_utils.escape_tag(key)
            value_tag = line_utils.escape_tag(tags[key])
            key_values.append(key_tag + u'=' + value_tag)
        key_values = u','.join(key_values)

        value_field = u'value={}'.format(value)
        value_meta_field = u'value_meta=' + value_meta_str

        data = key_values + u' ' + value_field + u',' + \
                            value_meta_field + u' ' + str(int(time_stamp))

        LOG.debug(data)

        return data, tenant_id



# -*- coding: utf-8 -*-
# Copyright 2017 FUJITSU LIMITED
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
import json
import os
from unittest import mock

from oslotest import base
from testtools import matchers

from monasca_persister.repositories.elasticsearch import events_repository
from monasca_persister.repositories import utils


class TestEvents(base.BaseTestCase):

    def __init__(self, *args, **kwds):
        super(TestEvents, self).__init__(*args, **kwds)
        self.events = None

    def setUp(self):
        super(TestEvents, self).setUp()

    def tearDown(self):
        super(TestEvents, self).tearDown()

    def test_parse_event(self):
        event = self._load_event('event_1')
        (project_id, timestamp, event_type, payload,
         dimensions), _ = utils.parse_events_message(event)
        self.assertEqual('de98fbff448f4f278a56e9929db70b03', project_id)
        self.assertEqual('2017-06-01 09:15:11.494606', timestamp)
        self.assertEqual('compute.instance.create.start', event_type)
        self.assertEqual('compute', dimensions['service'])
        self.assertEqual('notification.sample', dimensions['topic'])
        self.assertEqual('nova-compute:compute', dimensions['hostname'])
        self.assertIsNotNone(payload)
        self.assertThat(len(payload), matchers.GreaterThan(0))

    def test_normalize_timestamp(self):
        today = datetime.today().strftime('%Y-%m-%d')
        normalize_timestamp = events_repository.ElasticSearchEventsRepository._normalize_timestamp

        self.assertEqual(today, normalize_timestamp(None))
        self.assertEqual(today, normalize_timestamp(''))
        self.assertEqual(today, normalize_timestamp('foo'))
        self.assertEqual(today, normalize_timestamp('2017-02-3'))
        self.assertEqual(today, normalize_timestamp('2017-02-31'))

        self.assertEqual('2017-08-07', normalize_timestamp('2017-08-07 11:22:43'))

    @mock.patch('monasca_common.kafka.legacy_kafka_message')
    def _load_event(self, event_name, mock_kafka_message):
        if self.events is None:
            filepath = os.path.join(os.path.dirname(__file__), 'events.json')
            self.events = json.load(open(filepath))
        # create a kafka message envelope
        value = json.dumps(self.events[event_name])
        message = mock_kafka_message.LegacyKafkaMessage()
        message.value.return_value = value
        return message



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cassandra.concurrent import execute_concurrent_with_args
from oslo_config import cfg
from oslo_log import log
import simplejson as json

from monasca_persister.repositories.cassandra import abstract_repository
from monasca_persister.repositories.utils import parse_alarm_state_hist_message

LOG = log.getLogger(__name__)

UPSERT_CQL = (
    'update monasca.alarm_state_history USING TTL ? '
    'set metric = ?, old_state = ?, new_state = ?, sub_alarms = ?, reason = ?, reason_data = ? '
    'where tenant_id = ? and alarm_id = ? and time_stamp = ?')


class AlarmStateHistCassandraRepository(abstract_repository.AbstractCassandraRepository):
    def __init__(self):
        super(AlarmStateHistCassandraRepository, self).__init__()

        self._upsert_stmt = self._session.prepare(UPSERT_CQL)

    def process_message(self, message):
        (alarm_id, metrics, new_state, old_state, link, lifecycle_state, state_change_reason,
         sub_alarms_json_snake_case,
         tenant_id, time_stamp) = parse_alarm_state_hist_message(message)

        alarm_state_hist = (self._retention,
                            json.dumps(metrics, ensure_ascii=False).encode('utf8'),
                            old_state.encode('utf8'),
                            new_state.encode('utf8'),
                            sub_alarms_json_snake_case.encode('utf8'),
                            state_change_reason.encode('utf8'),
                            "{}".encode('utf8'),
                            tenant_id.encode('utf8'),
                            alarm_id.encode('utf8'),
                            time_stamp)

        return alarm_state_hist, tenant_id

    def write_batch(self, alarm_state_hists):
        while alarm_state_hists:
            num_rows = min(len(alarm_state_hists), cfg.CONF.kafka_alarm_history.batch_size)
            batch = alarm_state_hists[:num_rows]
            execute_concurrent_with_args(self._session, self._upsert_stmt, batch)
            alarm_state_hists = alarm_state_hists[num_rows:]



# -*- coding: utf-8 -*-
# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslotest import base

from monasca_persister.repositories.influxdb import line_utils

class TestInfluxdb(base.BaseTestCase):

    def setUp(self):
        super(TestInfluxdb, self).setUp()

    def tearDown(self):
        super(TestInfluxdb, self).tearDown()

    def test_line_utils_handles_utf8(self):
        utf8_name = u'ｎａｍｅ'
        self.assertEqual(u'"' + utf8_name + u'"', line_utils.escape_value(utf8_name))
        self.assertEqual(utf8_name, line_utils.escape_tag(utf8_name))

    def test_line_utils_escape_tag(self):
        simple = u"aaaaa"
        self.assertEqual(simple, line_utils.escape_tag(simple))
        complex = u"a\\ b,c="
        self.assertEqual("a\\\\\\ b\\,c\\=", line_utils.escape_tag(complex))

    def test_line_utils_escape_value(self):
        simple = u"aaaaa"
        self.assertEqual(u'"' + simple + u'"', line_utils.escape_value(simple))
        complex = u"a\\b\"\n"
        self.assertEqual(u"\"a\\\\b\\\"\\n\"", line_utils.escape_value(complex))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import signal
from unittest import mock

from oslo_config import cfg
from oslotest import base

CONF = cfg.CONF

NUMBER_OF_METRICS_PROCESSES = 2
NUMBER_OF_ALARM_HIST_PROCESSES = 3
NUMBER_OF_EVENTS_PROCESSES = 0


class FakeException(Exception):
    pass


class TestPersister(base.BaseTestCase):

    def setUp(self):
        super(TestPersister, self).setUp()

        from monasca_persister import persister
        self.persister = persister

        self._set_patchers()
        self._set_mocks()

    def _set_patchers(self):
        self.sys_exit_patcher = mock.patch.object(self.persister.sys, 'exit')
        self.log_patcher = mock.patch.object(self.persister, 'log')
        self.simport_patcher = mock.patch.object(self.persister, 'simport')
        self.cfg_patcher = mock.patch.object(self.persister, 'cfg')
        self.sleep_patcher = mock.patch.object(self.persister.time, 'sleep')
        self.process_patcher = mock.patch.object(
            self.persister.multiprocessing, 'Process')

    def _set_mocks(self):
        self.mock_sys_exit = self.sys_exit_patcher.start()
        self.mock_log = self.log_patcher.start()
        self.mock_simport = self.simport_patcher.start()
        self.mock_cfg = self.cfg_patcher.start()
        self.mock_sleep = self.sleep_patcher.start()
        self.mock_process_class = self.process_patcher.start()

        self.mock_cfg.CONF.kafka_metrics.num_processors = NUMBER_OF_METRICS_PROCESSES
        self.mock_cfg.CONF.kafka_alarm_history.num_processors = NUMBER_OF_ALARM_HIST_PROCESSES
        self.mock_cfg.CONF.kafka_events.num_processors = NUMBER_OF_EVENTS_PROCESSES

        self.mock_cfg.CONF.zookeeper = 'zookeeper'

        self.mock_sleep.side_effect = [FakeException, None]

        self.process_pool = _get_process_list(
            NUMBER_OF_METRICS_PROCESSES + NUMBER_OF_ALARM_HIST_PROCESSES + 1)
        self.mock_process_class.side_effect = self.process_pool

    def tearDown(self):
        super(TestPersister, self).tearDown()

        self.mock_sleep.side_effect = None

        self.mock_sys_exit.side_effect = None
        self.mock_sys_exit.reset_mock()

        self.sys_exit_patcher.stop()
        self.log_patcher.stop()
        self.simport_patcher.stop()
        self.cfg_patcher.stop()
        self.sleep_patcher.stop()
        self.process_patcher.stop()

        self.persister.processors = []
        self.persister.exiting = False

    def _run_persister(self):
        self.persister.main()
        self.assertTrue(self.mock_process_class.called)

    def _assert_clean_exit_handler_terminates_with_expected_signal(
            self, signum, expected):
        # in order to really exit, an exception is thrown

        self.mock_sys_exit.side_effect = FakeException
        self.assertRaises(FakeException, self.persister.clean_exit, signum)

        self.mock_sys_exit.assert_called_once_with(expected)

    def _assert_correct_number_of_processes_created(self):

        for p in self.process_pool[:-1]:
            self.assertTrue(p.start.called)

        self.assertFalse(self.process_pool[-1].called)

    def _assert_process_alive_status_called(self):

        for p in self.process_pool[:-1]:
            self.assertTrue(p.is_alive.called)

    def _assert_process_terminate_called(self):

        for p in self.process_pool[:-1]:
            self.assertTrue(p.terminate.called)

    def test_active_children_are_killed_during_exit(self):

        with mock.patch.object(self.persister.multiprocessing,
                               'active_children') as active_children,\
                mock.patch.object(self.persister.os, 'kill') as mock_kill:

            active_children.return_value = [mock.Mock(name='child-1', pid=1),
                                            mock.Mock(name='child-2', pid=2)]

            self.persister.clean_exit(0)

            mock_kill.assert_has_calls([mock.call(1, signal.SIGKILL),
                                        mock.call(2, signal.SIGKILL)])

    def test_active_children_kill_exception_is_ignored(self):

        with mock.patch.object(self.persister.multiprocessing,
                               'active_children') as active_children, \
                mock.patch.object(self.persister.os, 'kill') as mock_kill:

            active_children.return_value = [mock.Mock()]
            mock_kill.side_effect = FakeException

            self.persister.clean_exit(0)

            self.assertTrue(mock_kill.called)

    def test_clean_exit_does_nothing_when_exiting_is_true(self):
        self.persister.exiting = True

        self.assertIsNone(self.persister.clean_exit(0))

        self.assertFalse(self.mock_sys_exit.called)

    def test_exception_during_process_termination_is_ignored(self):
        dead_process = _get_process('raises_when_terminated')
        dead_process.terminate.side_effect = FakeException
        self.process_pool[0] = dead_process

        self._run_persister()

        self.assertTrue(dead_process.terminate.called)

    def test_if_not_sigterm_then_clean_exit_handler_terminates_with_signal(self):

        self._assert_clean_exit_handler_terminates_with_expected_signal(
            signal.SIGINT, signal.SIGINT)

    def test_if_sigterm_then_clean_exit_handler_terminates_with_zero(self):

        self._assert_clean_exit_handler_terminates_with_expected_signal(
            signal.SIGTERM, 0)

    def test_non_running_process_not_terminated(self):
        dead_process = _get_process('dead_process', is_alive=False)
        self.process_pool[0] = dead_process

        self._run_persister()

        self.assertTrue(dead_process.is_alive.called)
        self.assertFalse(dead_process.terminate.called)

    def test_running_processes_are_created_and_terminated(self):

        self._run_persister()

        self._assert_correct_number_of_processes_created()
        self._assert_process_alive_status_called()
        self._assert_process_terminate_called()

    def test_start_process_handler_creates_and_runs_persister(self):
        fake_kafka_config = mock.Mock()
        fake_repository = mock.Mock()

        with mock.patch('monasca_persister.kafka.legacy_kafka_persister.'
                        'LegacyKafkaPersister') as mock_persister_class:
            self.persister.start_process(fake_repository, fake_kafka_config)

            mock_persister_class.assert_called_once_with(
                fake_kafka_config, 'zookeeper', fake_repository)


def _get_process(name, is_alive=True):
    return mock.Mock(name=name, is_alive=mock.Mock(return_value=is_alive))


def _get_process_list(number_of_processes):
    processes = []
    for i in range(number_of_processes):
        processes.append(_get_process(name='process_{}'.format(i)))
    return processes



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslotest import base
from oslo_config import cfg

from monasca_persister.repositories.cassandra import alarm_state_history_repository
from monasca_persister.repositories.cassandra import connection_util
from monasca_persister.repositories import data_points


class TestAlarmStateHistoryRepo(base.BaseTestCase):
    def setUp(self):
        super(TestAlarmStateHistoryRepo, self).setUp()
        self._set_patchers()
        self._set_mocks()
        self.alarm_state_hist_repo = alarm_state_history_repository. \
            AlarmStateHistCassandraRepository()

    def tearDown(self):
        super(TestAlarmStateHistoryRepo, self).tearDown()

        self.mock_cfg.reset_mock()
        self.mock_conn_util_cluster.reset_mock()
        self.mock_conn_util_session.reset_mock()

        self.patch_cfg.stop()
        self.patch_conn_util_session.stop()
        self.patch_conn_util_cluster.stop()

    def _set_patchers(self):
        self.patch_cfg = mock.patch.object(
            alarm_state_history_repository.abstract_repository, 'conf')
        self.patch_conn_util_cluster = mock.patch.object(connection_util,
                                                         'create_cluster',
                                                         return_value=None)
        self.patch_conn_util_session =\
            mock.patch.object(connection_util, 'create_session',
                              return_value=mock.Mock(prepare=mock.Mock(
                                  return_value=None)))

    def _set_mocks(self):
        self.mock_cfg = self.patch_cfg.start()
        self.mock_conn_util_cluster = self.patch_conn_util_cluster.start()
        self.mock_conn_util_session = self.patch_conn_util_session.start()

    def test_process_message(self):
        message = mock.Mock()
        message.value.return_value = """{
            "alarm-transitioned": {
                "alarmId": "dummyid",
                "metrics": "dummymetrics",
                "newState": "dummynewState",
                "oldState": "dummyoldState",
                "link": "dummylink",
                "lifecycleState": "dummylifecycleState",
                "stateChangeReason": "dummystateChangeReason",
                "tenantId": "dummytenantId",
                "timestamp": "dummytimestamp",
                "subAlarms": {
                    "subAlarmExpression": "dummy_sub_alarm",
                    "currentValues": "dummy_values",
                    "metricDefinition": "dummy_definition",
                    "subAlarmState": "dummy_state"
                }
            }
        }"""
        self.alarm_state_hist_repo._retention = 0

        expected_output = [b'"sub_alarm_expression": "dummy_sub_alarm"',
                           b'"current_values": "dummy_values"',
                           b'"metric_definition": "dummy_definition"',
                           b'"sub_alarm_state": "dummy_state"']

        output, tenant_id = self.alarm_state_hist_repo.process_message(message)

        self.assertEqual(tenant_id, 'dummytenantId')
        self.assertEqual(output[0], self.alarm_state_hist_repo._retention)
        self.assertEqual(output[1], b'"dummymetrics"')
        self.assertEqual(output[2], b'dummyoldState')
        self.assertEqual(output[3], b'dummynewState')
        for elem in expected_output:
            self.assertIn(elem, output[4])
        self.assertEqual(output[5], b'dummystateChangeReason')
        self.assertEqual(output[6], b'{}')
        self.assertEqual(output[7], b'dummytenantId')
        self.assertEqual(output[8], b'dummyid')
        self.assertEqual(output[9], 'dummytimestamp')

    def test_write_batch(self):
        with mock.patch.object(alarm_state_history_repository,
                               'execute_concurrent_with_args',
                               return_value=0):
            cfg.CONF = mock.Mock(kafka_alarm_history=mock.Mock(batch_size=1))

            self._session, self._upsert_stmt = mock.Mock(), mock.Mock()
            alarm_state_hists_by_tenant = data_points.DataPointsAsList()
            alarm_state_hists_by_tenant.append('fake_tenant', 'elem')
            self.alarm_state_hist_repo.write_batch(alarm_state_hists_by_tenant)



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import influxdb
from influxdb.exceptions import InfluxDBClientError
from oslotest import base
from oslo_config import cfg
import six

from monasca_persister.repositories.influxdb.metrics_repository import MetricInfluxdbRepository


db_not_found = InfluxDBClientError(
    content='{"error": "database not found: db"}', code=404)

class TestMetricInfluxdbRepository(base.BaseTestCase):

    def setUp(self):
        super(TestMetricInfluxdbRepository, self).setUp()

    def tearDown(self):
        super(TestMetricInfluxdbRepository, self).tearDown()

    def _test_process_message(self, metrics_repo, data_points, metric, tenant):
        _dp, _tenant = metrics_repo.process_message(metric)
        self.assertIsInstance(_dp, six.string_types)
        self.assertEqual(_tenant, tenant)
        data_points.append(_tenant, _dp)

    @mock.patch.object(influxdb, 'InfluxDBClient')
    @mock.patch.object(cfg, 'CONF', return_value=None)
    def _test_write_batch(self, mock_conf, mock_influxdb_client,
                          db_per_tenant, db_exists, hours=0):
        mock_conf.influxdb.database_name = db_name = 'db'
        mock_conf.influxdb.db_per_tenant = db_per_tenant
        mock_conf.influxdb.default_retention_hours = hours
        t1 = u'fake_tenant_id_1'
        t2 = u'fake_tenant_id_2'
        m1 = self._get_metric(t1)
        m2 = self._get_metric(t2)
        metrics_repo = MetricInfluxdbRepository()
        data_points = metrics_repo.data_points_class()
        self._test_process_message(metrics_repo, data_points, m1, t1)
        self._test_process_message(metrics_repo, data_points, m2, t2)
        metrics_repo._influxdb_client = mock_influxdb_client

        if db_exists:
            metrics_repo._influxdb_client.write_points = mock.Mock()
        else:
            metrics_repo._influxdb_client.write_points = mock.Mock(
                side_effect=[db_not_found, None, db_not_found, None])
            rp = '{}h'.format(hours)
            if db_per_tenant:
                db1 = '%s_%s' % (db_name, t1)
                db2 = '%s_%s' % (db_name, t2)
                rp1 = mock.call(database=db1, default=True, name=rp,
                                duration=rp, replication='1')
                rp2 = mock.call(database=db2, default=True, name=rp,
                                duration=rp, replication='1')
                calls = [mock.call(db1), mock.call(db2)]
                rp_calls = [rp1, rp2]
            else:
                calls = [mock.call(db_name)]
                rp_calls = [mock.call(database=db_name, default=True,
                                      name=rp, duration=rp, replication='1')]
        metrics_repo.write_batch(data_points)
        if db_exists:
            mock_influxdb_client.create_database.assert_not_called()
            mock_influxdb_client.create_retention_policy.assert_not_called()
        else:
            mock_influxdb_client.create_database.assert_has_calls(
                calls, any_order=True)
            if hours > 0:
                mock_influxdb_client.create_retention_policy.assert_has_calls(
                    rp_calls, any_order=True)
            else:
                mock_influxdb_client.create_retention_policy.assert_not_called()

    def _get_metric(self, tenant_id):
        metric = '''
            {
              "metric":
              {
                "timestamp":1554725987408.1049804688,
                "name":"process.cpu_perc",
                "dimensions":
                {
                  "process_name":"persister",
                  "hostname":"devstack",
                  "service":"persister"
                },
                "value":7.5,
                "value_meta":null},
                "meta":
                {
                  "region":"RegionOne",
                  "tenantId":"''' + tenant_id + '''"
                },
                "creation_time":1554725988
            }
        '''
        message = mock.Mock()
        message.value.return_value = metric
        return message

    def test_write_batch_db_exists(self):
        self._test_write_batch(db_per_tenant=False, db_exists=True)

    def test_write_batch(self):
        self._test_write_batch(db_per_tenant=False, db_exists=False)

    def test_write_batch_db_exists_with_rp(self):
        self._test_write_batch(db_per_tenant=False, db_exists=True, hours=2016)

    def test_write_batch_with_rp(self):
        self._test_write_batch(db_per_tenant=False, db_exists=False, hours=2016)

    def test_write_batch_db_per_tenant_db_exists(self):
        self._test_write_batch(db_per_tenant=True, db_exists=True)

    def test_write_batch_db_per_tenant(self):
        self._test_write_batch(db_per_tenant=True, db_exists=False)

    def test_write_batch_db_per_tenant_db_exists_with_rp(self):
        self._test_write_batch(db_per_tenant=True, db_exists=True, hours=2016)

    def test_write_batch_db_per_tenant_with_rp(self):
        self._test_write_batch(db_per_tenant=True, db_exists=False, hours=2016)



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def escape_tag(tag):
    return str(tag).replace(
        "\\", "\\\\"
    ).replace(
        " ", "\\ "
    ).replace(
        ",", "\\,"
    ).replace(
        "=", "\\="
    )


def escape_value(value):
    return "\"{0}\"".format(
        str(value).replace(
            "\\", "\\\\"
        ).replace(
            "\"", "\\\""
        ).replace(
            "\n", "\\n"
        )
    )



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
# (C) Copyright 2017 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


kafka_common_opts = [
    cfg.StrOpt('consumer_id',
               help='Name/id of persister kafka consumer',
               advanced=True,
               default='monasca-persister'),
    cfg.StrOpt('client_id',
               help='id of persister kafka client',
               advanced=True,
               default='monasca-persister'),
    cfg.IntOpt('max_wait_time_seconds',
               help='Maximum wait time for write batch to database',
               default=30),
    cfg.IntOpt('fetch_size_bytes',
               help='Fetch size, in bytes. This value is set to the kafka-python defaults',
               default=4096),
    cfg.IntOpt('buffer_size',
               help='Buffer size, in bytes. This value is set to the kafka-python defaults',
               default='$kafka.fetch_size_bytes'),
    cfg.IntOpt('max_buffer_size',
               help='Maximum buffer size, in bytes, default value is 8 time buffer_size.'
                    'This value is set to the kafka-python defaults. ',
               default=32768),
    cfg.IntOpt('num_processors',
               help='Number of processes spawned by persister',
               default=1),
    cfg.BoolOpt('legacy_kafka_client_enabled',
                help='Enable legacy Kafka client. When set old version of '
                     'kafka-python library is used. Message format version '
                     'for the brokers should be set to 0.9.0.0 to avoid '
                     'performance issues until all consumers are upgraded.',
                default=True)
]

kafka_common_group = cfg.OptGroup(name='kafka',
                                  title='kafka')


def register_opts(conf):
    conf.register_group(kafka_common_group)
    conf.register_opts(kafka_common_opts, kafka_common_group)


def list_opts():
    return kafka_common_group, kafka_common_opts



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslotest import base

from monasca_persister.repositories.influxdb.alarm_state_history_repository \
    import AlarmStateHistInfluxdbRepository
from monasca_persister.repositories.influxdb import abstract_repository


class TestInfluxdbAlarmStateHistoryRepo(base.BaseTestCase):
    def setUp(self):
        super(TestInfluxdbAlarmStateHistoryRepo, self).setUp()
        with mock.patch.object(abstract_repository.cfg, 'CONF',
                               return_value=mock.Mock()):
            self.alarm_state_repo = AlarmStateHistInfluxdbRepository()

    def tearDown(self):
        super(TestInfluxdbAlarmStateHistoryRepo, self).tearDown()

    def test_process_message(self):
        message = mock.Mock()

        message.value.return_value = """{
            "alarm-transitioned": {
                "alarmId": "dummyid",
                "metrics": "dummymetrics",
                "newState": "dummynewState",
                "oldState": "dummyoldState",
                "link": "dummylink",
                "lifecycleState": "dummylifecycleState",
                "stateChangeReason": "dummystateChangeReason",
                "tenantId": "dummytenantId",
                "timestamp": "10",
                "subAlarms": {
                    "subAlarmExpression": "dummy_sub_alarm",
                    "currentValues": "dummy_values",
                    "metricDefinition": "dummy_definition",
                    "subAlarmState": "dummy_state"
                }
            }
        }"""
        expected_output = 'alarm_state_history,tenant_id=dummytenantId ' \
                          'tenant_id="dummytenantId",alarm_id="dummyid",' \
                          'metrics="\\"dummymetrics\\"",new_state="dummynewState"' \
                          ',old_state="dummyoldState",link="dummylink",' \
                          'lifecycle_state="dummylifecycleState",' \
                          'reason="dummystateChangeReason",reason_data="{}"'
        expected_dict = ['\\"sub_alarm_expression\\": \\"dummy_sub_alarm\\"',
                         '\\"metric_definition\\": \\"dummy_definition\\"',
                         '\\"sub_alarm_state\\": \\"dummy_state\\"',
                         '\\"current_values\\": \\"dummy_values\\"']

        actual_output, tenant_id = self.alarm_state_repo.process_message(message)

        self.assertEqual(tenant_id, 'dummytenantId')
        self.assertIn(expected_output, actual_output)
        for elem in expected_dict:
            self.assertIn(elem, actual_output)



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslotest import base
import simplejson as json

from monasca_persister.repositories import utils


class TestUtils(base.BaseTestCase):
    def setUp(self):
        super(TestUtils, self).setUp()

    def tearDown(self):
        super(TestUtils, self).tearDown()

    def test_parse_measurement_message(self):
        message = mock.Mock()
        message.value.return_value = """{
            "metric": {
              "name": "metric_name",
              "timestamp": "metric_timestamp",
              "value": "0.0",
              "value_meta": {

               },
              "dimensions": {}
          },
          "meta": {
              "region": "meta_region",
              "tenantId": "meta_tenantId"
          }
        }"""
        data = utils.parse_measurement_message(message)

        self.assertEqual(data[0], {})
        self.assertEqual(data[1], 'metric_name')
        self.assertEqual(data[2], 'meta_region')
        self.assertEqual(data[3], 'meta_tenantId')
        self.assertEqual(data[4], 'metric_timestamp')
        self.assertEqual(data[5], 0.0)
        self.assertEqual(data[6], {})

    def test_parse_alarm_state_hist_message(self):
        message = mock.Mock()
        message.value.return_value = """{
            "alarm-transitioned": {
                "alarmId": "dummyid",
                "metrics": "dummymetrics",
                "newState": "dummynewState",
                "oldState": "dummyoldState",
                "link": "dummylink",
                "lifecycleState": "dummylifecycleState",
                "stateChangeReason": "dummystateChangeReason",
                "tenantId": "dummytenantId",
                "timestamp": "dummytimestamp",
                "subAlarms": {
                    "subAlarmExpression": "dummy_sub_alarm",
                    "currentValues": "dummy_values",
                    "metricDefinition": "dummy_definition",
                    "subAlarmState": "dummy_state"
                }
            }
        }"""
        output = {'sub_alarm_expression': 'dummy_sub_alarm',
                  'current_values': 'dummy_values',
                  'metric_definition': 'dummy_definition',
                  'sub_alarm_state': 'dummy_state'}
        data = utils.parse_alarm_state_hist_message(message)
        self.assertEqual(data[0], 'dummyid')
        self.assertEqual(data[1], 'dummymetrics')
        self.assertEqual(data[2], 'dummynewState')
        self.assertEqual(data[3], 'dummyoldState')
        self.assertEqual(data[4], 'dummylink')
        self.assertEqual(data[5], 'dummylifecycleState')
        self.assertEqual(data[6], 'dummystateChangeReason')
        sub_alarms_data = json.loads(data[7])
        for elemKey, elemValue in output.items():
            self.assertIn(elemValue, sub_alarms_data[elemKey])
        self.assertEqual(data[8], 'dummytenantId')
        self.assertEqual(data[9], 'dummytimestamp')

    def test_parse_events_message(self):
        message = mock.Mock()
        message.value.return_value = """{
            "event": {
                "event_type": "dummy_event_type",
                "timestamp": "dummy_timestamp",
                "payload": "dummy_payload",
                "dimensions": "dummy_dimensions"
            },
            "meta": {
                "project_id": "dummy_project_id"
            }
        }"""

        (project_id, timestamp, event_type, payload,
         dimensions), _ = utils.parse_events_message(message)

        self.assertEqual(project_id, "dummy_project_id")
        self.assertEqual(timestamp, "dummy_timestamp")
        self.assertEqual(event_type, "dummy_event_type")
        self.assertEqual(payload, "dummy_payload")
        self.assertEqual(dimensions, "dummy_dimensions")



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from oslotest import base
from oslo_config import cfg

from monasca_common.kafka import consumer
from monasca_persister.kafka.legacy_kafka_persister import LegacyKafkaPersister
from monasca_persister.repositories import data_points
from monasca_persister.repositories.persister import LOG


class FakeException(Exception):
    pass


class TestPersisterRepo(base.BaseTestCase):
    def setUp(self):
        super(TestPersisterRepo, self).setUp()

        self._set_patchers()
        self._set_mocks()

        self.persister = LegacyKafkaPersister(self.mock_kafka,
                                              self.mock_zookeeper, mock.Mock())

    def _set_mocks(self):
        self.mock_kafka = mock.Mock()
        self.mock_kafka.topic = 'topic'
        self.mock_kafka.batch_size = 1
        self.mock_kafka.zookeeper_path = ''
        self.mock_kafka.group_id = 0
        self.mock_kafka.max_wait_time_seconds = 0
        self.mock_zookeeper = mock.Mock(uri='')

        self.mock_consumer_init = self.patch_kafka_init.start()
        self.mock_client_init = self.patch_kafka_client_init.start()
        self.mock_consumer_commit = self.patch_kafka_commit.start()
        self.mock_log_warning = self.patch_log_warning.start()
        self.mock_log_exception = self.patch_log_exception.start()

    def _set_patchers(self):
        self.patch_kafka_init = mock.patch.object(consumer.KafkaConsumer,
                                                  '__init__',
                                                  return_value=None)
        self.patch_kafka_commit = \
            mock.patch.object(consumer.KafkaConsumer, 'commit',
                              return_value=FakeException())
        self.patch_kafka_client_init = \
            mock.patch.object(consumer.kafka_client.KafkaClient, '__init__',
                              return_value=None)
        self.patch_log_warning = mock.patch.object(LOG, 'warning')
        self.patch_log_exception = mock.patch.object(LOG, 'exception')

    def tearDown(self):
        super(TestPersisterRepo, self).tearDown()

        self.mock_consumer_init.reset_mock()
        self.mock_client_init.reset_mock()
        self.mock_consumer_commit.reset_mock()
        self.mock_log_warning.reset_mock()
        self.mock_log_exception.reset_mock()

        self.patch_kafka_init.stop()
        self.patch_kafka_commit.stop()
        self.patch_kafka_client_init.stop()
        self.patch_log_warning.stop()
        self.patch_log_exception.stop()

    def test_flush_if_data_points_is_none(self):
        self.persister._data_points = None
        self.assertIsNone(self.persister._flush())

    def test_run_if_consumer_is_faulty(self):
        with mock.patch.object(os, '_exit', return_value=None) as mock_exit:
            self.persister._data_points = data_points.DataPointsAsDict()
            self.persister._consumer = mock.Mock(side_effect=FakeException)
            self.persister.run()
            mock_exit.assert_called_once_with(1)

    def test_run_logs_exception_from_consumer(self):
        with mock.patch.object(self.persister.repository, 'process_message',
                               side_effect=FakeException):
            self.persister._data_points = data_points.DataPointsAsDict()
            self.persister._consumer = ['aa']
            self.persister.run()
            self.mock_log_exception.assert_called()

    def test_run_commit_is_called_and_data_points_is_emptied(self):
        with mock.patch.object(self.persister.repository, 'process_message',
                               return_value=('message', 'tenant_id')):
            with mock.patch.object(self.persister, '_consumer',
                                   return_value=mock.Mock()) as mock_consumer:
                self.persister._data_points = data_points.DataPointsAsDict()
                self.persister._data_points.append('fake_tenant_id', 'some')
                self.persister._consumer.__iter__.return_value = ('aa', 'bb')
                self.persister._batch_size = 1
                self.persister.run()
                mock_consumer.commit.assert_called()
                self.assertEqual(0, self.persister._data_points.counter)

    def test_flush_logs_warning_and_exception(self):
        exception_msgs = ['partial write: points beyond retention policy dropped',
                          'unable to parse']
        with(mock.patch.object(cfg.CONF.repositories,
                               'ignore_parse_point_error', return_value=True)):
            for elem in exception_msgs:
                with mock.patch.object(LOG, 'info', side_effect=FakeException(
                        elem)):
                    self.persister._data_points = data_points.DataPointsAsDict()
                    self.persister._data_points.append('fake_tenant_id', 'some')
                    self.persister._flush()
                    self.mock_log_warning.assert_called()

    @mock.patch.object(LOG, 'info', side_effect=FakeException())
    def test_flush_logs_exception(self, mock_log_info):
        with(mock.patch.object(cfg.CONF.repositories,
                               'ignore_parse_point_error',
                               return_value=False)):
            mock_log_info.side_effect.message = 'some msg'
            self.persister._data_points = data_points.DataPointsAsDict()
            self.persister._data_points.append('fake_tenant_id', 'some')
            self.assertRaises(FakeException, self.persister._flush)
            self.mock_log_exception.assert_called()



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

repositories_opts = [
    cfg.StrOpt(
        name='metrics_driver',
        help='The repository driver to use for metrics',
        default=('monasca_persister.repositories.influxdb.metrics_repository:'
                 'MetricInfluxdbRepository')),
    cfg.StrOpt(
        name='alarm_state_history_driver',
        help='The repository driver to use for alarm state history',
        default=('monasca_persister.repositories.influxdb.'
                 'alarm_state_history_repository:'
                 'AlarmStateHistInfluxdbRepository')),
    cfg.StrOpt(
        name='events_driver',
        help='The repository driver to use for events',
        default=('monasca_persister.repositories.elasticsearch.events_repository:'
                 'ElasticSearchEventsRepository')),
    cfg.BoolOpt(
        'ignore_parse_point_error',
        help='Specifies if InfluxDB parse point errors should be ignored and measurements dropped',
        default=False)]

repositories_group = cfg.OptGroup(name='repositories',
                                  title='repositories')


def register_opts(conf):
    conf.register_group(repositories_group)
    conf.register_opts(repositories_opts, repositories_group)


def list_opts():
    return repositories_group, repositories_opts



# (C) Copyright 2019 Fujitsu Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from cassandra.policies import RetryPolicy
from oslotest import base

from monasca_persister.repositories.cassandra import retry_policy


class TestMonascaRetryPolicy(base.BaseTestCase):
    def setUp(self):
        super(TestMonascaRetryPolicy, self).setUp()
        self.monasca_retry_policy = retry_policy.MonascaRetryPolicy(0, 0, 0)

    def tearDown(self):
        super(TestMonascaRetryPolicy, self).tearDown()

    def test_on_read_timeout_when_retry_num_gt_read_attempts(self):
        self.monasca_retry_policy.read_attempts = 0
        retry_num = 1
        rethrow, none = self.monasca_retry_policy.on_read_timeout(mock.Mock(),
                                                                  mock.Mock(),
                                                                  mock.Mock(),
                                                                  0, 0,
                                                                  retry_num)
        self.assertEqual(rethrow, RetryPolicy.RETHROW)
        self.assertEqual(none, None)

    def test_on_read_timeout_when_responses_received_gt_required_and_data_retrieved_is_none(self):
        consistency = 'some_value'
        received_responses = 1
        required_responses = 0
        data_retrieved = None
        retry_num = 0
        self.monasca_retry_policy.read_attempts = 1
        returned_retry, returned_consistency = self.monasca_retry_policy. \
            on_read_timeout(mock.Mock(), consistency, required_responses,
                            received_responses, data_retrieved, retry_num)
        self.assertEqual(returned_retry, RetryPolicy.RETRY)
        self.assertEqual(consistency, returned_consistency)

    def test_on_read_timeout_if_previous_conditions_not_fulfilled(self):
        consistency = 'some_value'
        received_responses = 1
        required_reponses = 0
        data_retrieved = ['some']
        retry_num = 0
        self.monasca_retry_policy.read_attempts = 1
        returned_rethrow, returned_none = self.monasca_retry_policy. \
            on_read_timeout(mock.Mock(), consistency, required_reponses,
                            received_responses, data_retrieved, retry_num)
        self.assertEqual(returned_rethrow, RetryPolicy.RETHROW)
        self.assertEqual(returned_none, None)

    def test_on_write_timeout_if_retry_num_gt_or_eq_write_attempts(self):
        retry_num = 1
        self.monasca_retry_policy.write_attempts = 0
        returned_rethrow, returned_none = self.monasca_retry_policy. \
            on_write_timeout(mock.Mock(), mock.Mock(), mock.Mock(), 0, 0,
                             retry_num)
        self.assertEqual(returned_rethrow, RetryPolicy.RETHROW)
        self.assertEqual(returned_none, None)

    def test_on_write_timeout_if_retry_num_lt_write_attempts(self):
        retry_num = 0
        consistency = 0
        self.monasca_retry_policy.write_attempts = 1
        returned_retry, returned_consistency = self.monasca_retry_policy. \
            on_write_timeout(mock.Mock(), consistency, mock.Mock(), 0, 0,
                             retry_num)
        self.assertEqual(returned_retry, RetryPolicy.RETRY)
        self.assertEqual(returned_consistency, consistency)

    def test_on_unavailable_if_retry_num_lt_unavailable_attempts(self):
        consistency = 0
        retry_num = 0
        self.monasca_retry_policy.unavailable_attempts = 1

        returned_retry_next_host, returned_consistency = \
            self.monasca_retry_policy.on_unavailable(mock.Mock(),
                                                     consistency, 0, 0, retry_num)
        self.assertEqual(returned_consistency, consistency)
        self.assertEqual(returned_retry_next_host, RetryPolicy.RETRY_NEXT_HOST)

    def test_on_unavailable_if_retry_num_gt_or_eq_unavailable_attempts(self):
        consistency = 0
        retry_num = 1
        self.monasca_retry_policy.unavailable_attempts = 1

        returned_rethrow, returned_none = \
            self.monasca_retry_policy.on_unavailable(mock.Mock(),
                                                     consistency, 0, 0, retry_num)
        self.assertEqual(returned_none, None)
        self.assertEqual(returned_rethrow, RetryPolicy.RETHROW)



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import simplejson as json


def parse_measurement_message(message):

    decoded_message = json.loads(message.value())

    metric = decoded_message['metric']

    metric_name = metric['name']

    region = decoded_message['meta']['region']

    tenant_id = decoded_message['meta']['tenantId']

    time_stamp = metric['timestamp']

    value = float(metric['value'])

    value_meta = metric.get('value_meta', {})
    value_meta = {} if value_meta is None else value_meta

    return (metric.get('dimensions', {}), metric_name, region, tenant_id,
            time_stamp, value, value_meta)


def parse_alarm_state_hist_message(message):

    decoded_message = json.loads(message.value())

    alarm_transitioned = decoded_message['alarm-transitioned']

    alarm_id = alarm_transitioned['alarmId']

    metrics = alarm_transitioned['metrics']

    new_state = alarm_transitioned['newState']

    old_state = alarm_transitioned['oldState']

    # Key may not exist or value may be none, convert both to ""
    link = alarm_transitioned.get('link', "")
    link = "" if link is None else link

    # Key may not exist or value may be none, convert both to ""
    lifecycle_state = alarm_transitioned.get('lifecycleState', "")
    lifecycle_state = "" if lifecycle_state is None else lifecycle_state

    state_change_reason = alarm_transitioned['stateChangeReason']

    tenant_id = alarm_transitioned['tenantId']

    time_stamp = alarm_transitioned['timestamp']

    sub_alarms = alarm_transitioned['subAlarms']
    if sub_alarms:
        sub_alarms_json = json.dumps(sub_alarms, ensure_ascii=False)

        sub_alarms_json_snake_case = sub_alarms_json.replace(
            '"subAlarmExpression":',
            '"sub_alarm_expression":')

        sub_alarms_json_snake_case = sub_alarms_json_snake_case.replace(
            '"currentValues":',
            '"current_values":')

        # jobrs: I do not think that this shows up
        sub_alarms_json_snake_case = sub_alarms_json_snake_case.replace(
            '"metricDefinition":',
            '"metric_definition":')

        sub_alarms_json_snake_case = sub_alarms_json_snake_case.replace(
            '"subAlarmState":',
            '"sub_alarm_state":')
    else:
        sub_alarms_json_snake_case = "[]"

    return (alarm_id, metrics, new_state, old_state, link,
            lifecycle_state, state_change_reason,
            sub_alarms_json_snake_case, tenant_id, time_stamp)


def parse_events_message(message):

    decoded_message = json.loads(message.value())
    event_type = decoded_message['event']['event_type']
    timestamp = decoded_message['event']['timestamp']
    payload = decoded_message['event']['payload']
    project_id = decoded_message['meta']['project_id']
    dimensions = decoded_message['event']['dimensions']

    return (project_id, timestamp, event_type, payload, dimensions), project_id



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from oslo_config import cfg

from monasca_persister.conf import kafka_common
from monasca_persister.conf import types

kafka_metrics_group = cfg.OptGroup(name='kafka_metrics',
                                   title='kafka_metrics')
kafka_metrics_opts = [
    cfg.BoolOpt('enabled',
                help='Enable metrics persister',
                default=True),
    # NOTE(czarneckia) default by reference does not work with ListOpt
    cfg.ListOpt('uri',
                help='Comma separated list of Kafka broker host:port',
                default=['127.0.0.1:9092'],
                item_type=types.HostAddressPortType()),
    cfg.StrOpt('group_id',
               help='Kafka Group from which persister get data',
               default='1_metrics'),
    cfg.StrOpt('topic',
               help='Kafka Topic from which persister get data',
               default='metrics'),
    cfg.StrOpt('zookeeper_path',
               help='Path in zookeeper for kafka consumer group partitioning algorithm',
               default='/persister_partitions/$kafka_metrics.topic'),
    cfg.IntOpt('batch_size',
               help='Maximum number of metrics to buffer before writing to database',
               default=20000),
]

# Replace Default OPt with reference to kafka group option
kafka_common_opts = copy.deepcopy(kafka_common.kafka_common_opts)
for opt in kafka_common_opts:
    opt.default = '$kafka.{}'.format(opt.name)


def register_opts(conf):
    conf.register_group(kafka_metrics_group)
    conf.register_opts(kafka_metrics_opts + kafka_common_opts,
                       kafka_metrics_group)


def list_opts():
    return kafka_metrics_group, kafka_metrics_opts + kafka_common_opts
